# -*- coding: utf-8 -*-
"""
Created on Mon Jan 22 16:02:35 2024

@author: Jonathan
"""
from typing import Dict, Union
import os

from qtpy import QtWidgets, QtGui

ICONS_FOLDER = os.path.dirname(__file__)
standardIcon = QtWidgets.QApplication.style().standardIcon


def format_icon_path(name: str) -> str:
    return os.path.join(ICONS_FOLDER, name).replace("\\", "/")


def create_icon(name: str) -> QtGui.QIcon:
    return QtGui.QIcon(format_icon_path(name))


def create_pixmap(name: str) -> QtGui.QPixmap:
    return QtGui.QPixmap(format_icon_path(name))


icons: Dict[str, Union[QtGui.QIcon, QtGui.QPixmap, standardIcon]] = {
    'action': create_icon('action-icon.svg'),
    'add': create_icon('add-icon.svg'),
    'autolab': create_icon('autolab-icon.ico'),
    'autolab-pixmap': create_pixmap('autolab-icon.ico'),
    'config': create_icon('config-icon.svg'),
    'copy': create_icon('copy-icon.svg'),
    'disconnect': create_icon('disconnect-icon.svg'),
    'down': create_icon('down-icon.svg'),
    'export': create_icon('export-icon.svg'),
    'file': standardIcon(QtWidgets.QStyle.SP_FileDialogDetailedView),
    'folder': standardIcon(QtWidgets.QStyle.SP_DirIcon),
    'github': create_icon('github-icon.svg'),
    'import': create_icon('import-icon.svg'),
    'is-disable': create_icon('is-enable-icon.svg'),
    'is-enable': create_icon('is-enable-icon.svg'),
    'measure': create_icon('measure-icon.svg'),
    'monitor': create_icon('monitor-icon.svg'),
    'parameter': create_icon('parameter-icon.svg'),
    'paste': create_icon('paste-icon.svg'),
    'pdf': create_icon('pdf-icon.svg'),
    'plotter': create_icon('plotter-icon.svg'),
    'preference': create_icon('preference-icon.svg'),
    'read-save': create_icon('read-save-icon.svg'),
    'readthedocs': create_icon('readthedocs-icon.svg'),
    # 'recipe': create_icon('recipe-icon.svg'),
    'redo': create_icon('redo-icon.svg'),
    'reload': standardIcon(QtWidgets.QStyle.SP_BrowserReload),
    'remove': create_icon('remove-icon.svg'),
    'rename': create_icon('rename-icon.svg'),
    'scanner': create_icon('scanner-icon.svg'),
    'slider': create_icon('slider-icon.svg'),
    'undo': create_icon('undo-icon.svg'),
    'up': create_icon('up-icon.svg'),
    'variables': create_icon('variables-icon.svg'),
    'write': create_icon('write-icon.svg'),
}

icons.update({
    'int': create_icon('int-icon.svg'),
    'float': create_icon('float-icon.svg'),
    'str': create_icon('str-icon.svg'),
    'bytes': create_icon('bytes-icon.svg'),
    'bool': create_icon('bool-icon.svg'),
    'tuple': create_icon('tuple-icon.svg'),
    'ndarray': create_icon('ndarray-icon.svg'),
    'DataFrame': create_icon('dataframe-icon.svg'),
})



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 29 18:21:49 2019

@author: qchat
"""
from typing import Union
import time
import threading
from queue import Queue

import numpy as np
import pandas as pd
from qtpy import QtCore, QtWidgets

from ...variables import Variable
from ...elements import Variable as Variable_og


class MonitorManager:

    def __init__(self, gui: QtWidgets.QMainWindow):

        self.gui = gui

        # Configure a new thread
        self.thread = MonitorThread(self.gui.variable, self.gui.queue)
        self.thread.errorSignal.connect(self.error)

    def error(self, error: Exception):
        """ This function is called when the errorSignal of the thread is raised.
        It update the pause button and displays the error in the GUI """
        self.gui.pauseButton.setText('Resume')
        self.gui.setStatus(f'Error: {error} ', 10000, False)

    def start(self):
        """ This function start the thread """
        self.thread.start()

    def setDelay(self, value):
        """ Set the delay in the thread """
        self.thread.delay = value

    def getDelay(self):
        """ Returns the current delay of the thread """
        return self.thread.delay

    def isPaused(self):
        """ This function returns whether the thread is paused or not """
        return self.thread.pauseFlag.is_set()

    def resume(self):
        """ This function resume the monitoring """
        self.thread.pauseFlag.clear()

    def pause(self):
        """ This function pause the monitoring """
        self.thread.pauseFlag.set()

    def close(self):
        """ This function stops the thread and wait for its complete deletion """
        self.resume()
        self.thread.stopFlag.set()
        self.thread.wait()


class MonitorThread(QtCore.QThread):
    """ This thread class is dedicated to read the variable, and send its data to GUI through a queue """

    errorSignal = QtCore.Signal(object)

    def __init__(self, variable: Union[Variable, Variable_og], queue: Queue):

        super().__init__()
        self.variable = variable
        self.queue = queue

        self.pauseFlag = threading.Event()
        self.stopFlag = threading.Event()

        self.delay = 0

    def run(self):

        t_ini = time.time()
        pauseLength = 0
        pauseStartedTime = None

        while not self.stopFlag.is_set():
            # If the thread just resume, take into account the delay it has been paused
            if pauseStartedTime is not None:
                pauseLength += time.time() - pauseStartedTime
                pauseStartedTime = None

            # Time measure
            now = time.time() - t_ini - pauseLength

            try:
                # Measure variable
                value = self.variable()

                # Check type
                if not isinstance(value, (np.ndarray, pd.DataFrame)):  # should not float(array) because if 0D convert to float and loose information on type
                    try:
                        value = float(value)
                    except TypeError:
                        assert hasattr(value, "shape"), "If data is not a float, should be an array or a dataframe"

                # Send signal new data
                self.queue.put([now, value])

            except Exception as e:
                self.errorSignal.emit(e)
                self.pauseFlag.set()

            # If not the thread may be too fast
            time.sleep(self.delay)

            # pause
            while self.pauseFlag.is_set():
                if pauseStartedTime is None:
                    pauseStartedTime = time.time()
                time.sleep(0.1)



# -*- coding: utf-8 -*-
"""
Created on Mon Apr 17 23:23:51 2023

@author: jonathan
"""
from typing import Any, Union
import sys

import numpy as np
from qtpy import QtCore, QtWidgets

from .icons import icons
from .GUI_utilities import get_font_size, setLineEditBackground
from .GUI_instances import clearSlider
from ..config import get_control_center_config
from ..elements import Variable as Variable_og
from ..variables import Variable

if hasattr(QtCore.Qt.LeftButton, 'value'):
    LeftButton = QtCore.Qt.LeftButton.value
else:
    LeftButton = QtCore.Qt.LeftButton


class ProxyStyle(QtWidgets.QProxyStyle):
    """ https://stackoverflow.com/questions/67299834/pyqt-slider-not-come-to-a-specific-location-where-i-click-but-move-to-a-certain """
    def styleHint(self, hint, opt=None, widget=None, returnData=None):
        res = super().styleHint(hint, opt, widget, returnData)
        if hint == QtWidgets.QStyle.SH_Slider_AbsoluteSetButtons:
            res |= LeftButton
        return res


class Slider(QtWidgets.QMainWindow):

    changed = QtCore.Signal()  # Used by scanner to update filter when slider value changes

    def __init__(self,
                 variable: Union[Variable, Variable_og],
                 gui: QtWidgets.QMainWindow = None,
                 item: QtWidgets.QTreeWidgetItem = None):
        """ https://stackoverflow.com/questions/61717896/pyqt5-qslider-is-off-by-one-depending-on-which-direction-the-slider-is-moved """
        self.gui = gui  # gui can have setStatus, threadManager
        self.variable = variable
        self.item = item
        super().__init__()
        self.resize(self.minimumSizeHint())
        self.setWindowTitle(self.variable.address())
        self.setWindowIcon(icons['slider'])

        # Load configuration
        control_center_config = get_control_center_config()
        self.precision = int(float(control_center_config['precision']))

        self._font_size = get_font_size()

        # Slider
        self.slider_instantaneous = True
        self._last_moved = False  # Prevent double setting/readding after a slider has been moved with the slider_instantaneous=True

        if self.is_writable():
            self.true_min = self.variable.type(0)
            self.true_max = self.variable.type(10)
            self.true_step = self.variable.type(1)
        else:
            self.true_min = 0
            self.true_max = 10
            self.true_step = 1

        centralWidget = QtWidgets.QWidget()
        layoutWindow = QtWidgets.QVBoxLayout()
        layoutTopValue = QtWidgets.QHBoxLayout()
        layoutSlider = QtWidgets.QHBoxLayout()
        layoutBottomValues = QtWidgets.QHBoxLayout()

        centralWidget.setLayout(layoutWindow)
        layoutWindow.addLayout(layoutTopValue)
        layoutWindow.addLayout(layoutSlider)
        layoutWindow.addLayout(layoutBottomValues)

        self.instantCheckBox = QtWidgets.QCheckBox()
        self.instantCheckBox.setToolTip(
            "True: Changes instantaneously the value.\n" \
            "False: Changes the value when click released.")
        self.instantCheckBox.setCheckState(QtCore.Qt.Checked)
        self.instantCheckBox.stateChanged.connect(self.instantChanged)

        layoutTopValue.addWidget(QtWidgets.QLabel("Instant"))
        layoutTopValue.addWidget(self.instantCheckBox)

        self.valueWidget = QtWidgets.QLineEdit()
        self.valueWidget.setAlignment(QtCore.Qt.AlignCenter)
        self.valueWidget.setReadOnly(True)
        self.valueWidget.setText(f'{self.true_min}')
        setLineEditBackground(self.valueWidget, 'edited', self._font_size)

        layoutTopValue.addStretch()
        layoutTopValue.addWidget(QtWidgets.QLabel("Value"))
        layoutTopValue.addWidget(self.valueWidget)
        layoutTopValue.addStretch()
        layoutTopValue.addSpacing(40)

        self.sliderWidget = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.sliderWidget.setValue(0)
        self.sliderWidget.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self.sliderWidget.valueChanged.connect(self.valueChanged)
        self.sliderWidget.sliderReleased.connect(self.sliderReleased)
        self.sliderWidget.setStyle(ProxyStyle())

        button_minus = QtWidgets.QToolButton()
        button_minus.setArrowType(QtCore.Qt.LeftArrow)
        button_minus.clicked.connect(self.minusClicked)

        button_plus = QtWidgets.QToolButton()
        button_plus.setArrowType(QtCore.Qt.RightArrow)
        button_plus.clicked.connect(self.plusClicked)

        layoutSlider.addWidget(button_minus)
        layoutSlider.addWidget(self.sliderWidget)
        layoutSlider.addWidget(button_plus)

        self.minWidget = QtWidgets.QLineEdit()
        self.minWidget.setAlignment(QtCore.Qt.AlignLeft)
        self.minWidget.returnPressed.connect(self.minWidgetValueChanged)
        self.minWidget.textEdited.connect(lambda: setLineEditBackground(
            self.minWidget, 'edited', self._font_size))

        layoutBottomValues.addWidget(QtWidgets.QLabel("Min"))
        layoutBottomValues.addWidget(self.minWidget)
        layoutBottomValues.addSpacing(10)
        layoutBottomValues.addStretch()

        self.stepWidget = QtWidgets.QLineEdit()
        self.stepWidget.setAlignment(QtCore.Qt.AlignCenter)
        self.stepWidget.returnPressed.connect(self.stepWidgetValueChanged)
        self.stepWidget.textEdited.connect(lambda: setLineEditBackground(
            self.stepWidget, 'edited', self._font_size))

        layoutBottomValues.addWidget(QtWidgets.QLabel("Step"))
        layoutBottomValues.addWidget(self.stepWidget)
        layoutBottomValues.addStretch()
        layoutBottomValues.addSpacing(10)

        self.maxWidget = QtWidgets.QLineEdit()
        self.maxWidget.setAlignment(QtCore.Qt.AlignRight)
        self.maxWidget.returnPressed.connect(self.maxWidgetValueChanged)
        self.maxWidget.textEdited.connect(lambda: setLineEditBackground(
            self.maxWidget, 'edited', self._font_size))

        layoutBottomValues.addWidget(QtWidgets.QLabel("Max"))
        layoutBottomValues.addWidget(self.maxWidget)

        self.setCentralWidget(centralWidget)

        self.updateStep()

        self.resize(self.minimumSizeHint())

    def displayError(self, e: str):
        """ Wrapper to display errors """
        self.gui.setStatus(e, 10000, False) if self.gui and hasattr(
            self.gui, 'setStatus') else print(e, file=sys.stderr)

    def setVariableValue(self, value: Any):
        """ Wrapper to change variable value """
        if self.gui and hasattr(self.gui, 'threadManager') and self.item:
            self.gui.threadManager.start(
                self.item, 'write', value=value)
        else:
            self.variable(value)

    def is_writable(self):
        return self.variable.writable and self.variable.type in (int, float)

    def updateStep(self):

        if self.is_writable():
            slider_points = 1 + int(
                np.floor((self.true_max - self.true_min) / self.true_step))
            self.true_max = self.variable.type(
                self.true_step*(slider_points - 1) + self.true_min)

            self.minWidget.setText(f'{self.true_min}')
            setLineEditBackground(self.minWidget, 'synced', self._font_size)
            self.maxWidget.setText(f'{self.true_max}')
            setLineEditBackground(self.maxWidget, 'synced', self._font_size)
            self.stepWidget.setText(f'{self.true_step}')
            setLineEditBackground(self.stepWidget, 'synced', self._font_size)

            temp = self.slider_instantaneous
            self.slider_instantaneous = False
            self.sliderWidget.setMinimum(0)
            self.sliderWidget.setSingleStep(1)
            self.sliderWidget.setTickInterval(1)
            self.sliderWidget.setMaximum(slider_points - 1)
            self.slider_instantaneous = temp
        else: self.badType()

    def updateTrueValue(self, old_true_value: Any):

        if self.is_writable():
            new_cursor_step = round(
                (old_true_value - self.true_min) / self.true_step)
            slider_points = 1 + int(
                np.floor((self.true_max - self.true_min) / self.true_step))
            if new_cursor_step > (slider_points - 1):
                new_cursor_step = slider_points - 1
            elif new_cursor_step < 0:
                new_cursor_step = 0

            temp = self.slider_instantaneous
            self.slider_instantaneous = False
            self.sliderWidget.setSliderPosition(new_cursor_step)
            self.slider_instantaneous = temp

            true_value = self.variable.type(
                new_cursor_step*self.true_step + self.true_min)
            self.valueWidget.setText(f'{true_value:.{self.precision}g}')
            setLineEditBackground(self.valueWidget, 'edited', self._font_size)
        else: self.badType()

    def stepWidgetValueChanged(self):

        if self.is_writable():
            old_true_value = self.variable.type(self.valueWidget.text())
            try:
                true_step = self.variable.type(self.stepWidget.text())
                assert true_step != 0, "Can't have step=0"
                self.true_step = true_step
            except Exception as e:
                e = f"Variable {self.variable.address()}: {e}"
                self.displayError(e)
            else:
                self.updateStep()
                self.updateTrueValue(old_true_value)
        else: self.badType()

    def minWidgetValueChanged(self):

        if self.is_writable():
            old_true_value = self.variable.type(self.valueWidget.text())
            try:
                self.true_min = self.variable.type(self.minWidget.text())
            except Exception as e:
                e = f"Variable {self.variable.address()}: {e}"
                self.displayError(e)
            else:
                self.updateStep()
                self.updateTrueValue(old_true_value)
        else: self.badType()

    def maxWidgetValueChanged(self):

        if self.is_writable():
            old_true_value = self.variable.type(self.valueWidget.text())
            try:
                self.true_max = self.variable.type(self.maxWidget.text())
            except Exception as e:
                e = f"Variable {self.variable.address()}: {e}"
                self.displayError(e)
            else:
                self.updateStep()
                self.updateTrueValue(old_true_value)
        else: self.badType()

    def sliderReleased(self):
        """ Do something when the cursor is released """
        if self.is_writable():
            if self.slider_instantaneous and self._last_moved:
                self._last_moved = False
                return None
            self._last_moved = False
            value = self.sliderWidget.value()
            true_value = self.variable.type(
                value*self.true_step + self.true_min)
            self.valueWidget.setText(f'{true_value:.{self.precision}g}')
            setLineEditBackground(self.valueWidget, 'synced', self._font_size)
            self.setVariableValue(true_value)
            self.changed.emit()
            self.updateStep()
        else:
            self.badType()

    def valueChanged(self, value: Any):
        """ Do something with the slider value when the cursor is moved """
        if self.is_writable():
            self._last_moved = True
            true_value = self.variable.type(
                value*self.true_step + self.true_min)
            self.valueWidget.setText(f'{true_value:.{self.precision}g}')
            if self.slider_instantaneous:
                setLineEditBackground(self.valueWidget, 'synced', self._font_size)
                self.setVariableValue(true_value)
                self.changed.emit()
            else:
                setLineEditBackground(self.valueWidget, 'edited', self._font_size)
            # self.updateStep()  # Don't use it here, infinite loop leading to crash if set min > max
        else: self.badType()

    def instantChanged(self, value):
        self.slider_instantaneous = self.instantCheckBox.isChecked()

    def minusClicked(self):
        self.sliderWidget.setSliderPosition(self.sliderWidget.value()-1)
        self._last_moved = False
        if not self.slider_instantaneous: self.sliderReleased()

    def plusClicked(self):
        self.sliderWidget.setSliderPosition(self.sliderWidget.value()+1)
        self._last_moved = False
        if not self.slider_instantaneous: self.sliderReleased()

    def badType(self):
        setLineEditBackground(self.valueWidget, 'edited', self._font_size)
        setLineEditBackground(self.minWidget, 'edited', self._font_size)
        setLineEditBackground(self.stepWidget, 'edited', self._font_size)
        setLineEditBackground(self.maxWidget, 'edited', self._font_size)

        e = f"Variable {self.variable.address()}: Variable should be writable int or float"
        self.displayError(e)

    def closeEvent(self, event):
        """ This function does some steps before the window is really killed """
        clearSlider(self.variable)

        if self.gui is None:
            QtWidgets.QApplication.quit()  # close the slider app



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Fri May 17 15:04:04 2019

@author: quentin.chateiller
"""

import sys
import argparse
from typing import List, Type

import autolab


def print_help():
    print(f'Autolab {autolab.__version__}')
    print()
    print('Usage:')
    print('  autolab [-h] <command> ...')
    print()
    print('Commands:')
    print('  gui                   Start the Graphical User Interface')
    print('  plotter               Start the Plotter')
    print('  add_device            Start add device menu')
    print('  install_drivers       Install drivers from GitHub')
    print('  driver                Driver interface')
    print('  device                Device interface')
    print('  doc                   Open the online documentation (readthedocs)')
    print('  report                Open the online report/suggestions webpage (github)')
    print('  infos                 Displays the available drivers and devices configuration')
    print()
    print('General Options:')
    print('  -h, --help            Show this help message')
    print()


def main():

    args = sys.argv

    # No command provided or -h/--help option : print help
    if (len(args) == 1) or (args[1] == '-h') or (args[1] == '--help'):
        print_help()

    else:
        command = args[1]   # first is 'autolab'

        # Update the sys.argv for parsers
        args = [f'autolab {command}'] + args[2: ]  # first is 'autolab' and second is command
        sys.argv = args

        # Removed bellow and similar because getattr will get every standard command (only difference is now it raises error if gives too much arguments)
        # if command == 'gui':
        #     autolab.gui()
        if command == 'driver':
            driver_parser(args)
        elif command == 'device':
            device_parser(args)
        elif command in dir(autolab):  # Execute autolab.command if exists
            attr = getattr(autolab, command)
            if hasattr(attr, '__call__'):
                attr(*args[1: ])
            else:
                print(attr)
        else:
            print(f"Command {command} not known. Autolab doesn't have Super Cow Power... yet ^^")

    sys.exit()


####################################################################################
######################### autolab driver/device utilities ##########################
def process_config(args_list: List[str]):

    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("-D", "--driver", type=str, dest="driver", help="Set the nickname or driver to use: 1) uses nickname if it is defined in devices_config.ini OR(if it is not) 2) Set the driver name to use." )
    parser.add_argument("-C", "--connection", type=str, dest="connection", help="Set the connection type to use for the connection." )
    parser.add_argument("-A", "--address", type=str, dest="address", help="Set the address to use for the communication." )
    parser.add_argument("-O", "--other", nargs='+', dest="other", help="Set other parameters [ports (e.g SOCKET), board_index, slots,...)." )

    args, unknown = parser.parse_known_args(args_list)

    config = {}
    if args.connection is not None: config['connection'] = args.connection
    if args.address is not None: config['address'] = args.address
    if args.other is not None:
        for part in args.other:
            part = part.replace(' ', '')
            config[part.split('=')[0]] = part.split('=')[1]

    # Help for autolab driver/device -h/--help and autolab driver/device -h/--help -D driver_name
    if not args.driver or (
            (not args.driver
             and args_list[1] == '-h') or args_list[1] == '--help'):
        print_help_parser(parser,args_list); sys.exit()
    if args.driver and (args_list[1] == '-h' or args_list[1] == '--help'):
        if args.driver in autolab._devices.list_devices():
            try:
                args.driver = autolab._devices.get_final_device_config(args.driver)["driver"]
            except:
                pass
        assert args.driver in autolab._drivers.list_drivers(), f"Driver {args.driver} not known"
        autolab.config_help(args.driver,_parser=True)
        sys.exit()

    return args.driver, config, parser


def print_help_parser(parser: argparse.ArgumentParser, args_list: List[str]):
    parser.usage = f"""

----------------  General informations:  ----------------

This is a very basic help message for usage of {args_list[0]}. More info can be found on read the doc website (command: autolab doc). To display a more extensive help, please have a look at the section help below.

    Usage:   {args_list[0]} -D driver_name -C connection -A address -h

Recquired connection arguments (capital letters):
    -D driver_name: name of the driver to use (e.g.: agilent_33220A). driver_name can be either the driver_name or the defined nickname, as defined by the user in the devices_config.ini. See below for accessing the list of available drivers.
    -C connection type to use to communicate with the device (e.g.: VISA, VXI11, SOCKET, TELNET, USB, GPIB, ...). You may access the available connections types with an help (see below helps section).
    -A address: full address to reach the device that depends on the connection type (e.g.: 192.168.0.2  [for VXI11]) and on the way you configured the instrument.


----------------  Helps (-h option)  ----------------

Three helps are configured:
    {args_list[0]} -h
    Print this help message.

    {args_list[0]} -h -D driver_name
    Print useful informations on the {args_list[0].split(' ')[1]}, including available connection types, etc.

    {args_list[0]} -D driver_name -C connection -A address -h
    Full help message about the {args_list[0].split(' ')[1]}. This requires the address to be valid and the {args_list[0].split(' ')[1]} to be instantiated.


----------------  List of available drivers and local configurations (-D option) ----------------

To display the full list of available drivers and local configurations, use:
    autolab infos


---------------- Accepted options ----------------"""

    parser.print_help()
########################## autolab driver/device utilities ##########################
#####################################################################################


#####################################################################################
################################## autolab driver ###################################
def driver_parser(args_list: List[str]):

    # autolab driver -D mydummy -C CONN -A GPIB0::5::INSTR -m set_verbose(1)
    # autolab driver -D mydummy -C CONN -A GPIB0::5::INSTR -verb 1

    # Reading of connection information
    driver_name, config, parser = process_config(args_list)

    # Reading of methods
    parser.add_argument("-m", "--methods", nargs='+', dest="methods", help="Set the methods to use." )
    parser.add_argument("-h", "--help", dest="help", help="Print this help message." )

    # Instantiation of driver.py and driver_utilities.py
    global driver_instance
    assert 'connection' in config, f"Must provide a connection for the driver using -C connection with connection being for this driver among {autolab._drivers.get_connection_names(autolab._drivers.load_driver_lib(driver_name))}"
    driver_instance = autolab.get_driver(driver_name, **config)

    if driver_name in autolab._config.list_all_devices_configs():
        # Load config object
        config = dict(autolab._config.get_device_config(driver_name))
        # Check if driver provided
        assert 'driver' in config, f"Driver name not found in driver config '{driver_name}'"
        driver_name = config['driver']

    driver_utilities = autolab._drivers.load_driver_utilities_lib(driver_name + '_utilities')
    driver_utilities_instance = driver_utilities.Driver_parser(driver_instance, driver_name)

    # Add arguments to the existing parser (driver dependant)
    parser = driver_utilities_instance.add_parser_arguments(parser)

    # Add help and usage to the parser only if "-h" options requested
    if '-h' in args_list:
        dirver_infos_for_usage = build_driver_infos_for_usage(driver_name, driver_instance)
        parser.usage = driver_utilities_instance.add_parser_usage(dirver_infos_for_usage)
        parser.usage = parser.usage + """
    autolab driver -D nickname -m 'some_methods1(arg1,arg2=23)' 'some_methods2(arg1="test")'
    Execute some_methods of the driver. A list of available methods is present at the top of this help along with arguments definition. Note if strings are passed as arguments, mind the usage of ' and " (follow some_methods2 example)."""
        parser.print_help()
        sys.exit()

    args = parser.parse_args()

    # Finally, execute functions according to the arguments provided
    driver_utilities_instance.do_something(args)

    # Process given methods
    if args.methods:
        global message
        method_list = [method_args[0] for method_args in autolab._drivers.get_instance_methods(driver_instance)]
        for method in args.methods:
            message = None
            print(method)
            assert method.split('(')[0] in method_list, f"Method {method} not known or bound. Methods known are: {method_list}"
            print(f'\nExecuting command:  {method}')
            exec(f"message = driver_instance.{method}", globals())
            if message is not None: print(f'Return:  {message}\n')

    # Driver closing
    driver_utilities_instance.exit()

def build_driver_infos_for_usage(driver_name: str, driver_instance: Type):
    driver_lib = autolab._drivers.load_driver_lib(driver_name)

    # build necessary list
    list_connection_classes = autolab._drivers.get_connection_names(driver_lib)
    list_additional_classes = autolab._drivers.get_module_names(driver_lib)
    list_instance_methods_and_args = autolab._drivers.get_instance_methods(driver_instance)

    mess = '\n----------------  Driver informations:  ----------------\n'
    mess += '\n    Available connections types (-C option):\n'
    for connection in list_connection_classes:
        mess += f'     - {connection}\n'
    mess += '\n'
    mess += '\n    Available additional modules:\n'
    if list_additional_classes==[]:
        mess += f'     {list_additional_classes}'
    else:
        for additional_class in list_additional_classes:
            mess += f'     - {additional_class}\n'
    mess += '\n'
    mess += '\n    Available methods (with arguments):'
    for method in list_instance_methods_and_args:
        mess += f'\n     - {method[0]}({",".join(method[1])})'
    mess += '\n'

    return mess
################################## autolab driver ###################################
#####################################################################################


#####################################################################################
################################## autolab device ###################################
def device_parser(args_list: List[str]):

    # autolab device -D mydummy -e amplitude -p C:\Users\               GET AND SAVE VARIABLE VALUE
    # autolab device -D mydummy -e something                            EXECUTE ACTION
    # autolab device -D mydummy -e amplitude -v 4                       SET VARIABLE VALUE
    # autolab device -D mydummy -e channel1.amplitude -h                DISPLAY ELEMENT HELP

    # Reading of connection information
    driver_name, config, parser = process_config(args_list)

    # Parser configuration
    parser.add_argument('-e', '--element', type=str, dest="element", help='Address of the element to open' )
    parser.add_argument("-v", "--value", type=str, dest="value", help='Value to set')
    parser.add_argument("-p", "--path", type=str, dest="path", help='Path where to save data')
    parser.add_argument("-h", "--help", action='store_true', dest="help", help='Display element help')

    # In autolab driver, this is done after the last help request
    if "-e" in list(set([x for x in args_list if args_list.count(x) >= 2])):
        print("Warning, device will only consider the last element provided with -e. Use driver instead to do multiple operations in the same command")

    args, unknown = parser.parse_known_args(args_list)

    # Instantiation
    instance = autolab.get_device(driver_name, **config)
    element = instance

    # Open element
    if args.element is not None:
        for name in args.element.split('.'):
            element = getattr(element, name)

    # Execute order
    if args.help: element.help()

    elif args.path is not None:
        assert element._element_type == 'variable', "This element is not a Variable"
        value = element()
        element.save(args.path,value=value)

    elif args.value is not None:
        assert element._element_type in ('variable', 'action'), "Please provide a Variable or Action element using -e <element>"
        if element._element_type == 'variable': element(args.value)
        elif element._element_type == 'action': element(args.value)
    else:
        if element._element_type == 'variable': print(element())
        elif element._element_type == 'action': element()

    args = parser.parse_args()  # return error if give bad arg

    instance.close()
################################## autolab device ###################################
#####################################################################################


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
Created on Oct 2022

@author: jonathan based on qchat
"""

from typing import List, Union, Any
import os
import sys
import csv

import numpy as np
import pandas as pd

try:
    from pandas._libs.lib import no_default
except:
    no_default = None

from qtpy import QtWidgets

from ...paths import PATHS
from ...config import load_config
from ...utilities import data_to_dataframe, SUPPORTED_EXTENSION
from ...devices import DEVICES, get_element_by_address, list_devices
from ...elements import Variable as Variable_og
from ...variables import list_variables, get_variable, Variable


def find_delimiter(filename):
    sniffer = csv.Sniffer()
    with open(filename) as fp:
        try:
            text = fp.read(5000)
            if text.startswith("#"):
                text = text[len(text.split("\n")[0])+len("\n"):]
            delimiter = sniffer.sniff(text).delimiter
        except:
            # delimiter = ","  # only 1 column
            delimiter = no_default
        if delimiter in ("e", "."):  # sniffer got it wrong
            delimiter = no_default
    return delimiter


def _skiprows(filename):
    with open(filename) as fp:
        if fp.read(1) not in ("#", "!", "\n"):
            skiprows = None
        else:
            skiprows = 1
            fp.readline()
            while fp.read(1) in ("#", "!", "\n"):
                skiprows += 1
                fp.readline()
    return skiprows


def find_header(filename, sep=no_default, skiprows=None):
    try:
        df = pd.read_csv(filename, sep=sep, header=None, nrows=5, skiprows=skiprows)
    except Exception:
        if type(skiprows) is not None: skiprows += 1
        df = pd.read_csv(filename, sep=sep, header=None, nrows=5, skiprows=skiprows)
    else:
        if skiprows == 1:
            try:
                df_columns = pd.read_csv(filename, sep=sep, header="infer",
                                         skiprows=0, nrows=0)
            except Exception:
                pass
            else:
                columns = list(df_columns.columns)
                if columns[0] == "#":
                    columns.pop(0)
                    if len(columns) == len(df):
                        return (0, 1, columns)  # (header, skiprows, columns)

    try:
        first_row = df.iloc[0].values.astype("float")
        return (("infer", skiprows, no_default)
                if tuple(first_row) == tuple([i for i in range(len(first_row))])
                else (None, skiprows, no_default))
    except:
        pass
    df_header = pd.read_csv(filename, sep=sep, nrows=5, skiprows=skiprows)

    return (("infer", skiprows, no_default)
            if tuple(df.dtypes) != tuple(df_header.dtypes)
            else (None, skiprows, no_default))


def importData(filename):
    """ This function open the data with the provided filename """

    skiprows = _skiprows(filename)
    sep = find_delimiter(filename)
    (header, skiprows, columns) = find_header(filename, sep, skiprows)
    try:
        data = pd.read_csv(filename, sep=sep, header=header,
                           skiprows=skiprows, names=columns)
    except TypeError:
        data = pd.read_csv(filename, sep=sep, header=header,
                           skiprows=skiprows, names=None)  # for pandas 1.2: names=None but sep=no_default
    except:
        data = pd.read_csv(filename, sep="\t", header=header,
                           skiprows=skiprows, names=columns)

    assert len(data) != 0, "Can't import empty DataFrame"
    data = data_to_dataframe(data)
    return data


class DataManager:

    def __init__(self, gui: QtWidgets.QMainWindow):

        self.gui = gui
        self._clear()
        self.overwriteData = True

        plotter_config = load_config("plotter_config")
        if ('device' in plotter_config.sections()
                and 'address' in plotter_config['device']):
            self.variable_address = str(plotter_config['device']['address'])
        else:
            self.variable_address = ''

    def _clear(self):
        self.datasets = []
        self.last_variables = []

    def setOverwriteData(self, value: bool):
        self.overwriteData = bool(value)

    def getDatasetsNames(self) -> List[str]:
        names = []
        for dataset in self.datasets:
            names.append(dataset.name)
        return names

    @staticmethod
    def getUniqueName(name: str, names_list: List[str]):
        """ This function adds a number next to basename in case this basename
        is already taken """
        raw_name, extension = os.path.splitext(name)

        if extension in (".txt", ".csv", ".dat"):
            basename = raw_name
            putext = True
        else:
            basename = name
            putext = False

        compt = 0
        while True:
            if name in names_list:
                compt += 1
                if putext:
                    name = basename+'_' + str(compt) + extension
                else:
                    name = basename + '_' + str(compt)
            else:
                break
        return name

    def set_variable_address(self, value: str):
        """ This function set the address of the target variable """
        # Can raise errors
        self.getVariable(value)
        # If no errors
        self.variable_address = value

    def get_variable_address(self) -> str:
        """ This function returns the value of the target device value """
        return self.variable_address

    def getVariable(self, name: str) -> Union[Variable, Variable_og]:
        """ This function returns the name of the target device value """
        assert name != '', 'Need to provide a variable name'
        device = name.split(".")[0]

        if device in list_variables():
            assert device == name, f"Can only use '{device}' directly"
            variable = get_variable(name)
        elif device in list_devices():
            assert device in DEVICES, f"Device '{device}' is closed"
            variable = get_element_by_address(name)  # Can raise 'name' not found in module 'device'
            assert isinstance(variable, Variable_og), (
                f"Need a variable but '{name}' is a {str(type(variable).__name__)}")
            assert variable.readable, f"Variable '{name}' is not readable"
            var_type = str(variable.type).split("'")[1]
            assert variable.type in (int, float, np.ndarray, pd.DataFrame), (
                f"Datatype '{var_type}' of '{name}' is not supported")
        else:
            assert False, f"'{device}' is neither a device nor a variable"

        return variable

    def data_comboBoxClicked(self):
        """ This function select a dataset """
        if len(self.datasets) == 0:
            self.gui.save_pushButton.setEnabled(False)
            self.gui.clear_pushButton.setEnabled(False)
            self.gui.clear_all_pushButton.setEnabled(False)
        else:
            self.updateDisplayableResults()

    def importActionClicked(self):
        """ This function prompts the user for a dataset filename,
        and import the dataset"""
        filenames = QtWidgets.QFileDialog.getOpenFileNames(
            self.gui, "Import data file", PATHS['last_folder'],
            filter=SUPPORTED_EXTENSION)[0]
        if not filenames:
            return None
        else:
            self.importAction(filenames)

    def importAction(self, filenames: List[str]):
        dataset = None
        for i, filename in enumerate(filenames):
            try:
                dataset = self.importData(filename)
            except Exception as e:
                self.gui.setStatus(
                    f"Impossible to load data from {filename}: {e}",
                    10000, False)
                if len(filenames) != 1:
                    print(f"Impossible to load data from {filename}: {e}",
                          file=sys.stderr)
            else:
                self.gui.setStatus(f"File {filename} loaded successfully", 5000)
        if dataset is not None:
            self.gui.figureManager.start(dataset)

        path = os.path.dirname(filename)
        PATHS['last_folder'] = path

    def importDeviceData(self, variable: Union[Variable, Variable_og, pd.DataFrame, Any]):
        """ This function open the data of the provided device """
        if isinstance(variable, pd.DataFrame):
            name = variable.name if hasattr(variable, 'name') else 'dataframe'
            data = variable
        elif isinstance(variable, (Variable, Variable_og)):
            name = variable.address()
            data = variable()  # read value
        else:
            name = 'data'
            data = variable
        data = data_to_dataframe(data)  # format value

        if self.overwriteData:
            data_name = name
        else:
            names_list = self.getDatasetsNames()
            data_name = DataManager.getUniqueName(
                name, names_list)

        dataset = self.newDataset(data_name, data)
        return dataset

    def importData(self, filename: str):
        """ This function open the data with the provided filename """
        # OPTIMIZE: could add option to choose in GUI all options
        data = importData(filename)
        name = os.path.basename(filename)

        if self.overwriteData:
            data_name = name
        else:
            names_list = self.getDatasetsNames()
            data_name = DataManager.getUniqueName(name, names_list)

        dataset = self.newDataset(data_name, data)
        return dataset

    def _addData(self, new_dataset):
        names = self.gui.dataManager.getDatasetsNames()

        if self.gui.overwriteDataButton.isChecked() and new_dataset.name in names:
            dataSet_id = names.index(new_dataset.name) + 1
            current_dataset = self.gui.dataManager.datasets[dataSet_id-1]

            if not new_dataset.data.equals(current_dataset.data):
                current_dataset.update(new_dataset)
        else:
            # Prepare a new dataset in the plotter
            self.gui.dataManager.addDataset(new_dataset)

    def getData(self, nbDataset: int, var_list: List[str], selectedData: int = 0):
        """ This function returns to the figure manager the required data """
        dataList = []
        for i in range(selectedData, nbDataset+selectedData):
            if i < len(self.datasets):
                dataset = self.datasets[-(i+1)]
                try:
                    data = dataset.getData(var_list)
                except:
                    data = None
                dataList.append(data)
            else:
                break
        dataList.reverse()

        return dataList

    def saveButtonClicked(self):
        """ This function is called when the save button is clicked.
        It asks a path and starts the procedure to save the data """
        dataset = self.getLastSelectedDataset()
        if dataset is not None:
            filename = QtWidgets.QFileDialog.getSaveFileName(
                self.gui, caption="Save data", directory=PATHS['last_folder'],
                filter=SUPPORTED_EXTENSION)[0]
            path = os.path.dirname(filename)

            if path != '':
                PATHS['last_folder'] = path
                self.gui.setStatus('Saving data...', 5000)

                dataset.save(filename)
                self.gui.figureManager.save(filename)

                self.gui.setStatus(
                    f'Last dataset successfully saved in {filename}', 5000)

    def clear(self):
        """ Clear displayed dataset """
        data_name = self.gui.data_comboBox.currentText()
        index = self.gui.data_comboBox.currentIndex()
        nbr_data = self.gui.data_comboBox.count()  # how many item left

        try:
            dataset = self.getLastSelectedDataset()
            self.deleteData(dataset)
            self.gui.data_comboBox.removeItem(index)
            self.gui.setStatus(f"Removed {data_name}", 5000)
        except Exception as e:
            self.gui.setStatus(f"Can't delete: {e}", 10000, False)
            pass

        if self.gui.data_comboBox.count() == 0:
            self.clear_all()
            return

        else:
            # if last point but exist other data takes previous data else keep index
            if index == (nbr_data-1) and index != 0:
                index -= 1

            self.gui.data_comboBox.setCurrentIndex(index)

        self.data_comboBoxClicked()
        self.gui.plugin_refresh()

    def clear_all(self):
        """ This reset any recorded data, and the GUI accordingly """
        self._clear()

        self.gui.figureManager.clearData()
        self.gui.variable_x_comboBox.clear()
        self.gui.variable_y_comboBox.clear()
        self.gui.data_comboBox.clear()

        self.data_comboBoxClicked()
        self.gui.plugin_refresh()

    def deleteData(self, dataset):
        """ This function remove dataset from the datasets"""
        self.datasets.remove(dataset)

    def getLastSelectedDataset(self):
        """ This return the current (last selected) dataset """
        if len(self.datasets) > 0:
            return self.datasets[self.gui.data_comboBox.currentIndex()]

    def addDataset(self, dataset):
        """ This function add the given dataset to datasets list """
        self.datasets.append(dataset)

    def newDataset(self, name: str, data: pd.DataFrame):
        """ This function creates a new dataset """
        dataset = Dataset(name, data)
        self._addData(dataset)
        return dataset

    def updateDisplayableResults(self):
        """ This function update the combobox in the GUI that displays the
        names of the results that can be plotted """
        dataset = self.getLastSelectedDataset()

        variables_list = list(dataset.data.columns)

        if variables_list != self.last_variables:
            self.last_variables = variables_list
            result_names = []

            for result_name in variables_list:
                    try:
                        float(dataset.data.iloc[0][result_name])
                        result_names.append(result_name)
                    except Exception as e:
                        self.gui.setStatus(f"Can't plot data: {e}", 10000, False)
                        return None

            variable_x = self.gui.variable_x_comboBox.currentText()
            variable_y = self.gui.variable_y_comboBox.currentText()

            # If can, put back previous x and y variable in combobox
            is_id = (len(variables_list) != 0
                     and variables_list[0] == "id"
                     and len(variables_list) > 2)

            if is_id:
                # move id form begin to end
                name=result_names.pop(0)
                result_names.append(name)

            AllItems = [self.gui.variable_x_comboBox.itemText(i)
                        for i in range(self.gui.variable_x_comboBox.count())]

            # only refresh if change labels, to avoid gui refresh that prevent user to click on combobox
            if result_names != AllItems:
                self.gui.variable_x_comboBox.clear()
                self.gui.variable_y_comboBox.clear()

                self.gui.variable_x_comboBox.addItems(result_names)  # slow (0.25s)

                if len(result_names) != 0:
                    # move first item to end (only for y axis)
                    name=result_names.pop(0)
                    result_names.append(name)

                self.gui.variable_y_comboBox.addItems(result_names)

                if variable_x in variables_list:
                    index = self.gui.variable_x_comboBox.findText(variable_x)
                    self.gui.variable_x_comboBox.setCurrentIndex(index)
                # else:  # BAD IDEA: Too slow to be useful, better to rearange items like before
                #     self.gui.variable_x_comboBox.setCurrentIndex(0+is_id)

                if variable_y in variables_list:
                    index = self.gui.variable_y_comboBox.findText(variable_y)
                    self.gui.variable_y_comboBox.setCurrentIndex(index)
                # else:
                #     self.gui.variable_y_comboBox.setCurrentIndex(1+is_id)
        else:
            self.gui.figureManager.reloadData()  # 0.1s


class Dataset():

    def __init__(self, name: str, data: pd.DataFrame):

        self.name = name
        self.data = data

    def getData(self, var_list: List[str]):
        """ This function returns a dataframe with two columns:
        the parameter value, and the requested result value """
        if var_list[0] == var_list[1]: return self.data.loc[:, [var_list[0]]]
        else: return self.data.loc[:, var_list]

    def update(self, dataset):
        """ Change name and data of this dataset """
        self.name = dataset.name
        self.data = dataset.data

    def save(self,filename: str):
        """ This function saved the dataset in the provided path """
        self.data.to_csv(filename, index=False)

    def __len__(self):
        """ Returns the number of data point of this dataset """
        return len(self.data)



# -*- coding: utf-8 -*-
"""
Created on Oct 2022

@author: jonathan based on qchat
"""
import os
import sys
import queue
import time
import uuid
from typing import Type, Union, Any

import pandas as pd
from qtpy import QtCore, QtWidgets, uic, QtGui

from .figure import FigureManager
from .data import DataManager
from .thread import ThreadManager
from .treewidgets import TreeWidgetItemModule
from ..icons import icons
from ..GUI_utilities import get_font_size, setLineEditBackground, MyLineEdit
from ..GUI_instances import clearPlotter, closePlotter
from ...devices import list_devices
from ...elements import Variable as Variable_og
from ...variables import Variable
from ...config import load_config


class MyQTreeWidget(QtWidgets.QTreeWidget):

    def __init__(self, gui, parent=None):
        self.gui = gui
        super().__init__(parent)

        self.setAcceptDrops(True)

    def dropEvent(self, event):
        """ This function is used to add a plugin to the plotter """
        plugin_name = event.source().last_drag
        if isinstance(plugin_name, str):
            self.gui.addPlugin(plugin_name, plugin_name)
        self.setGraphicsEffect(None)

    def dragEnterEvent(self, event):

        if (event.source() is self) or (
                hasattr(event.source(), "last_drag")
                    and isinstance(event.source().last_drag, str)):
            event.accept()
            shadow = QtWidgets.QGraphicsDropShadowEffect(
                blurRadius=25, xOffset=3, yOffset=3)
            self.setGraphicsEffect(shadow)
        else:
            event.ignore()

    def dragLeaveEvent(self, event):
        self.setGraphicsEffect(None)

    def keyPressEvent(self, event):
        if (event.key() == QtCore.Qt.Key_C
                and event.modifiers() == QtCore.Qt.ControlModifier):
            self.copy_item(event)
        else:
            super().keyPressEvent(event)

    def copy_item(self, event):
        if len(self.selectedItems()) == 0:
            super().keyPressEvent(event)
            return None
        item = self.selectedItems()[0]  # assume can select only one item
        if hasattr(item, 'variable'):
            text = item.variable.address()
        elif hasattr(item, 'action'):
            text = item.action.address()
        elif hasattr(item, 'module'):
            if hasattr(item.module, 'address'):
                text = item.module.address()
            else:
                text = item.name
        else:
            print(f'Should not be possible: {item}')
            super().keyPressEvent(event)
            return None

        # Copy the text to the system clipboard
        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setText(text)

class Plotter(QtWidgets.QMainWindow):

    def __init__(self, has_parent: bool = False):

        self.active = False
        self.has_parent = has_parent  # Only for closeEvent
        self.all_plugin_list = []
        self.active_plugin_dict = {}

        self._font_size = get_font_size()

        # Configuration of the window
        super().__init__()
        ui_path = os.path.join(os.path.dirname(__file__), 'interface.ui')
        uic.loadUi(ui_path, self)
        self.setWindowTitle("AUTOLAB - Plotter")
        self.setWindowIcon(icons['plotter'])

        # Loading of the different centers
        self.figureManager = FigureManager(self)
        self.dataManager = DataManager(self)

        self.threadManager = ThreadManager(self)
        self.threadDeviceDict = {}
        self.threadItemDict = {}

        # Save button
        self.save_pushButton.clicked.connect(self.dataManager.saveButtonClicked)
        self.save_pushButton.setEnabled(False)

        # Clear button
        self.clear_pushButton.clicked.connect(self.dataManager.clear)
        self.clear_all_pushButton.clicked.connect(self.dataManager.clear_all)
        self.clear_pushButton.setEnabled(False)
        self.clear_all_pushButton.setEnabled(False)

        # Open button
        self.openButton.clicked.connect(self.dataManager.importActionClicked)

        # comboBox with data id
        self.data_comboBox.activated.connect(self.dataManager.data_comboBoxClicked)

        # Number of traces
        self.nbTraces_lineEdit.setText(f'{self.figureManager.nbtraces:g}')
        self.nbTraces_lineEdit.returnPressed.connect(self.nbTracesChanged)
        self.nbTraces_lineEdit.textEdited.connect(lambda: setLineEditBackground(
            self.nbTraces_lineEdit,'edited', self._font_size))
        setLineEditBackground(self.nbTraces_lineEdit, 'synced', self._font_size)

        self.variable_x_comboBox.currentIndexChanged.connect(
            self.axisChanged)
        self.variable_y_comboBox.currentIndexChanged.connect(
            self.axisChanged)

        # Variable frame
        self.variable_lineEdit = MyLineEdit()
        self.variable_lineEdit.skip_has_eval = True
        self.variable_lineEdit.use_np_pd = False
        self.variable_lineEdit.setToolTip('Variable address e.g. ct400.scan.data')
        self.layout_variable.addWidget(self.variable_lineEdit, 1, 2)
        self.variable_lineEdit.setText(f'{self.dataManager.variable_address}')
        self.variable_lineEdit.returnPressed.connect(self.variableChanged)
        self.variable_lineEdit.textEdited.connect(lambda: setLineEditBackground(
            self.variable_lineEdit, 'edited', self._font_size))
        setLineEditBackground(self.variable_lineEdit, 'synced', self._font_size)

        # Plot button
        self.plotDataButton.clicked.connect(lambda state: self.refreshPlotData())

        # Timer
        self.timer_time = 0.5  # This plotter is not meant for fast plotting like the monitor, be aware it may crash with too high refreshing rate
        self.timer = QtCore.QTimer(self)
        self.timer.setInterval(int(self.timer_time*1000)) # ms
        self.timer.timeout.connect(self.autoRefreshPlotData)

        self.auto_plotDataButton.clicked.connect(self.autoRefreshChanged)

        # Delay
        self.delay_lineEdit.setText(str(self.timer_time))
        self.delay_lineEdit.returnPressed.connect(self.delayChanged)
        self.delay_lineEdit.textEdited.connect(lambda: setLineEditBackground(
            self.delay_lineEdit, 'edited', self._font_size))
        setLineEditBackground(self.delay_lineEdit, 'synced', self._font_size)
        # / Variable frame

        self.overwriteDataButton.clicked.connect(self.overwriteDataChanged)

        self.setAcceptDrops(True)

        # timer to load plugin to tree
        self.timerPlugin = QtCore.QTimer(self)
        self.timerPlugin.setInterval(50) # ms
        self.timerPlugin.timeout.connect(self.timerAction)

        # queue and timer to add/remove plot from plugin
        self.queue_driver = queue.Queue()
        self.dict_widget = {}
        self.timerQueue = QtCore.QTimer(self)
        self.timerQueue.setInterval(int(50)) # ms
        self.timerQueue.timeout.connect(self._queueDriverHandler)
        self._stop_timerQueue = False

        self.processPlugin()

        self.splitter.setSizes([600, 100])  # height
        self.splitter_2.setSizes([310, 310, 310])  # width

        for splitter in (self.splitter, self.splitter_2, self.splitter_3):
            for i in range(splitter.count()):
                handle = splitter.handle(i)
                handle.setStyleSheet("background-color: #DDDDDD;")
                handle.installEventFilter(self)

    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.Enter:
            obj.setStyleSheet("background-color: #AAAAAA;")  # Hover color
        elif event.type() == QtCore.QEvent.Leave:
            obj.setStyleSheet("background-color: #DDDDDD;")  # Normal color
        return super().eventFilter(obj, event)

    def createWidget(self, widget: Type, *args, **kwargs):
        """ Function used by a driver to add a widget.
        Mainly used to open a figure outside the GUI from a driver. """
        unique_name = str(uuid.uuid4())
        self.queue_driver.put(('create', unique_name, widget, args, kwargs))

        start = time.time()
        while True:
            widget_created = self.dict_widget.get(unique_name)

            if widget_created: return widget_created

            time.sleep(0.01)
            if (time.time() - start) > 1:
                print(f"Warning: Importation of {widget} too long, skip it",
                      file=sys.stderr)
                return None

    def removeWidget(self, widget: Type):
        """ Function used by a driver to remove a widget record from GUI """
        self.queue_driver.put(('remove', "", widget, (), {}))

    def _queueDriverHandler(self):
        """ Adds/Removes plot from a driver requests throught a queue """
        while not self.queue_driver.empty():
            data = self.queue_driver.get()
            action = data[0]
            widget_name = data[1]
            widget = data[2]
            args = data[3]
            kwargs = data[4]

            if action == 'create':
                widget = widget(*args, **kwargs)
                self.dict_widget[widget_name] = widget
                try: self.figureManager.ax.addItem(widget)
                except Exception as e: self.setStatus(str(e), 10000, False)
            elif action == "remove":
                d = self.dict_widget
                if widget is not None:
                    widget_pos = list(d.values()).index(widget)
                    if widget_pos is not None:
                        widget_name = list(d)[widget_pos]
                        widget = d.get(widget_name)
                        if widget is not None:
                            widget = d.pop(widget_name)
                            try: self.figureManager.ax.removeItem(widget)
                            except Exception as e:
                                self.setStatus(str(e), 10000, False)

        if self._stop_timerQueue:
            self.timerQueue.stop()
            self._stop_timerQueue = False

    def timerAction(self):
        """ This function checks if a module has been loaded and put to the queue.
        If so, associate item and module """
        threadItemDictTemp = self.threadItemDict.copy()
        threadDeviceDictTemp = self.threadDeviceDict.copy()

        for item_id in threadDeviceDictTemp:

            item = threadItemDictTemp[item_id]
            module = threadDeviceDictTemp[item_id]

            self.associate(item, module)
            item.setExpanded(True)

            self.threadItemDict.pop(item_id)
            self.threadDeviceDict.pop(item_id)

        if len(threadItemDictTemp) == 0:
            self.timerPlugin.stop()

    def itemClicked(self, item):
        """ Function called when a normal click has been detected in the tree.
            Check the association if it is a main item """
        if (item.parent() is None
            and item.loaded is False
            and id(item) not in self.threadItemDict):
            self.threadItemDict[id(item)] = item  # needed before start of timer to avoid bad timing and to stop thread before loading is done
            self.threadManager.start(item, 'load')  # load device and add it to queue for timer to associate it later (doesn't block gui while device is openning)
            self.timerPlugin.start()
            self.timerQueue.start()

    def rightClick(self, position):
        """ Function called when a right click has been detected in the tree """
        item = self.tree.itemAt(position)
        if hasattr(item,'menu'):
            item.menu(position)

    def hide_plugin_frame(self):
        sizes = self.splitter_3.sizes()
        self.splitter_3.setSizes([0, sizes[0]])

    def processPlugin(self):
        # Create frame
        frame = QtWidgets.QFrame()
        self.splitter_3.insertWidget(0, frame)
        frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        layout = QtWidgets.QVBoxLayout(frame)

        frame2 = QtWidgets.QFrame()
        layout2 = QtWidgets.QHBoxLayout(frame2)
        layout2.setContentsMargins(0,0,0,0)

        label = QtWidgets.QLabel('Plugin:', frame)
        label.setToolTip("Drag and drop a device from the control panel to add a plugin to the plugin tree")
        layout2.addWidget(label)
        font = QtGui.QFont()
        font.setBold(True)
        label.setFont(font)

        hide_plugin_button = QtWidgets.QPushButton('-')
        hide_plugin_button.setMaximumSize(40, 23)
        hide_plugin_button.clicked.connect(self.hide_plugin_frame)
        layout2.addWidget(hide_plugin_button)

        layout.addWidget(frame2)

        # Tree widget configuration
        self.tree = MyQTreeWidget(self, frame)
        layout.addWidget(self.tree)
        self.tree.setHeaderLabels(['Plugin', 'Type', 'Actions', 'Values', ''])
        self.tree.header().setDefaultAlignment(QtCore.Qt.AlignCenter)
        self.tree.header().setMinimumSectionSize(15)
        self.tree.header().resizeSection(0, 160)
        self.tree.header().hideSection(1)
        self.tree.header().resizeSection(2, 50)
        self.tree.header().resizeSection(3, 70)
        self.tree.header().resizeSection(4, 15)
        self.tree.header().setStretchLastSection(False)
        self.tree.setAlternatingRowColors(True)

        self.tree.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.tree.itemClicked.connect(self.itemClicked)
        self.tree.customContextMenuRequested.connect(self.rightClick)

        plotter_config = load_config("plotter_config")

        self.splitter_3.setSizes([280, 800])

        if ('plugin' in plotter_config.sections()
                and len(plotter_config['plugin']) != 0):
            for plugin_nickname, plugin_name in plotter_config['plugin'].items():
                self.addPlugin(plugin_name, plugin_nickname)

    def addPlugin(self, plugin_name: str, plugin_nickname: str):

        if plugin_name in list_devices():
            plugin_nickname = self.getUniqueName(plugin_nickname)
            self.all_plugin_list.append(plugin_nickname)
            item = TreeWidgetItemModule(self.tree,plugin_name,plugin_nickname,self)
            item.setBackground(0, QtGui.QColor('#9EB7F5'))  # blue
        else:
            self.setStatus(
                f"Error: plugin {plugin_name} not found in devices_config.ini",
                10000, False)

    def associate(self, item, module):

        item.load(module)
        self.active_plugin_dict[item.nickname] = module

        read_init_list = module._read_init_list
        for mod in module._mod.values():
            read_init_list.extend(mod._read_init_list)
        for variable in read_init_list:
            try:
                variable()
            except:
                self.setStatus(
                    f"Can't read variable {variable.address()} on instantiation",
                    10000, False)
        try:
            data = self.dataManager.getLastSelectedDataset().data
            data = data[[self.variable_x_comboBox.currentText(),
                         self.variable_y_comboBox.currentText()]].copy()
            module.instance.refresh(data)
        except Exception:
            pass

    def getUniqueName(self, basename: str):
        """ This function adds a number next to basename in case this basename
        is already taken """
        names = self.all_plugin_list
        name = basename

        compt = 0
        while True:
            if name in names:
                compt += 1
                name = basename+'_'+str(compt)
            else:
                break
        return name

    def dropEvent(self, event):
        """ Import data from filenames dropped """
        filenames = [e.toLocalFile() for e in event.mimeData().urls()]
        self.dataManager.importAction(filenames)

        qwidget_children = self.findChildren(QtWidgets.QWidget)
        for widget in qwidget_children:
            widget.setGraphicsEffect(None)

    def dragEnterEvent(self, event):
        """ Check that drop filenames """
        # only accept if there is at least one filename in the dropped filenames -> refuse folders
        if (event.mimeData().hasUrls()
            and any([os.path.isfile(e.toLocalFile())
                     for e in event.mimeData().urls()])):
            event.accept()

            qwidget_children = self.findChildren(QtWidgets.QWidget)
            for widget in qwidget_children:
                shadow = QtWidgets.QGraphicsColorizeEffect()
                shadow.setColor(QtGui.QColor(20,20,20))
                shadow.setStrength(1)
                widget.setGraphicsEffect(shadow)
        else:
            event.ignore()

    def dragLeaveEvent(self, event):
        qwidget_children = self.findChildren(QtWidgets.QWidget)
        for widget in qwidget_children:
            widget.setGraphicsEffect(None)

    def plugin_refresh(self):
        if self.active_plugin_dict:
            self.clearStatus()
            dataset = self.dataManager.getLastSelectedDataset()
            if hasattr(dataset, "data"):
                data = dataset.data
                variable_x = self.variable_x_comboBox.currentText()
                variable_y = self.variable_y_comboBox.currentText()
                if (variable_x in data.columns and variable_y in data.columns):
                    data = data[[variable_x, variable_y]].copy()
                else:
                    data = None
            else:
                data = None

            for module in self.active_plugin_dict.values():
                if hasattr(module.instance, "refresh"):
                    try:
                        module.instance.refresh(data)
                    except Exception as e:
                        self.setStatus(f"Error in plugin {module.name}: {e}",
                                       10000, False)

    def overwriteDataChanged(self):
        """ Set overwrite name for data import """
        self.dataManager.setOverwriteData(self.overwriteDataButton.isChecked())

    def autoRefreshChanged(self):
        """ Set if auto refresh call for device data """
        if self.auto_plotDataButton.isChecked():
            self.timer.start()
        else:
            self.timer.stop()

    def autoRefreshPlotData(self):
        """ Function that refresh plot every timer interval """
        # OPTIMIZE: timer should not call a heavy function, idealy just take data to plot
        self.refreshPlotData()

    def refreshPlotData(self, variable: Union[Variable, Variable_og, pd.DataFrame, Any] = None):
        """ This function get the last dataset data and display it onto the Plotter GUI """
        try:
            if variable is None:
                variable_address = self.dataManager.get_variable_address()
                variable = self.dataManager.getVariable(variable_address)
            dataset = self.dataManager.importDeviceData(variable)
            self.figureManager.start(dataset)
            self.setStatus(f"Display the data: '{dataset.name}'", 5000)
        except Exception as e:
            self.setStatus(f"Can't refresh data: {e}", 10000, False)

    def variableChanged(self):
        """ This function start the update of the target value in the data manager
        when a changed has been detected """
        # Send the new value
        try:
            value = str(self.variable_lineEdit.text())
            self.dataManager.set_variable_address(value)
        except Exception as e:
            self.setStatus(f"Can't change variable name: {e}", 10000, False)
        else:
            # Rewrite the GUI with the current value
            self.updateDeviceValueGui()

    def updateDeviceValueGui(self):
        """ This function ask the current value of the target value in the data
        manager, and then update the GUI """
        value = self.dataManager.get_variable_address()
        self.variable_lineEdit.setText(f'{value}')
        setLineEditBackground(self.variable_lineEdit, 'synced', self._font_size)

    def axisChanged(self, index):
        """ This function is called when the displayed result has been changed in
        the combo box. It proceeds to the change. """
        self.figureManager.clearData()

        if (self.variable_x_comboBox.currentIndex() != -1
                and self.variable_y_comboBox.currentIndex() != -1):
            self.figureManager.reloadData()

    def nbTracesChanged(self):
        """ This function is called when the number of traces displayed has been changed
        in the GUI. It proceeds to the change and update the plot. """
        value = self.nbTraces_lineEdit.text()

        check = False
        try:
            value = int(float(value))
            assert value > 0
            self.figureManager.nbtraces = value
            check = True
        except:
            pass

        self.nbTraces_lineEdit.setText(f'{self.figureManager.nbtraces:g}')
        setLineEditBackground(self.nbTraces_lineEdit, 'synced', self._font_size)

        if check is True and self.variable_y_comboBox.currentIndex() != -1:
            self.figureManager.reloadData()

    def closeEvent(self, event):
        """ This function does some steps before the window is closed (not killed) """
        if hasattr(self, 'timer'): self.timer.stop()
        self.timerPlugin.stop()
        self.timerQueue.stop()

        clearPlotter()

        if not self.has_parent:
            closePlotter()

        super().closeEvent(event)

        if not self.has_parent:
            QtWidgets.QApplication.quit()  # close the plotter app

    def close(self):
        """ This function does some steps before the window is killed """
        for children in self.findChildren(QtWidgets.QWidget):
            children.deleteLater()

        super().close()

    def delayChanged(self):
        """ This function start the update of the delay in the thread manager
        when a changed has been detected """
        # Send the new value
        try:
            value = float(self.delay_lineEdit.text())
            assert value >= 0
            self.timer_time = value
        except:
            pass

        # Rewrite the GUI with the current value
        self.updateDelayGui()

    def updateDelayGui(self):
        """ This function ask the current value of the delay in the data
        manager, and then update the GUI """

        value = self.timer_time
        self.delay_lineEdit.setText(f'{value:g}')
        self.timer.setInterval(int(value*1000))  # ms
        setLineEditBackground(self.delay_lineEdit, 'synced', self._font_size)

    def setStatus(self, message: str, timeout: int = 0, stdout: bool = True):
        """ Modify the message displayed in the status bar and add error message to logger """
        self.statusBar.showMessage(message, timeout)
        if not stdout: print(message, file=sys.stderr)

    def clearStatus(self):
        """ Erase the message displayed in the status bar """
        self.setStatus('')



# -*- coding: utf-8 -*-
"""
Created on Thu Jun 13 10:25:49 2019

@author: quentin.chateiller
"""

from typing import List, Union

from .drivers import get_driver_path, get_driver
from .config import list_all_devices_configs, get_device_config
from .elements import Module, Element

# Storage of the devices
DEVICES = {}

# After DEVICES to avoid circular import
from .variables import update_allowed_dict

# =============================================================================
# DEVICE CLASS
# =============================================================================

class Device(Module):

    def __init__(self, device_name: str, instance, device_config: dict):
        """ device_config is returned by :meth:`get_final_device_config` """

        self.device_config = device_config  # hidden from completion
        self.driver_path = get_driver_path(device_config["driver"])

        super().__init__(None, {'name': device_name, 'object': instance,
                                'help': f'Device {device_name} at {self.driver_path}'})

    def close(self):
        """ This function close the connection of the current physical device """
        # Remove read and write signals from gui
        try:
            # condition avoid reopenning connection if use close twice
            if self.name in DEVICES:
                for struc in self.get_structure():
                    element = get_element_by_address(struc[0])
                    if struc[1] == 'variable':
                        element._read_signal = None
                        element._write_signal = None
                    if struc[1] == 'action':
                        element._write_signal = None
        except: pass

        try: self.instance.close()
        except: pass

        del DEVICES[self.name]
        update_allowed_dict()

    def __dir__(self):
        """ For auto-completion """
        return (self.list_modules() + self.list_variables()
                + self.list_actions() + ['driver_path', 'close', 'help', 'instance'])


# =============================================================================
# DEVICE GET FUNCTION
# =============================================================================

def get_element_by_address(address: str) -> Element:
    """ Returns the Element located at the provided address """
    address_list = address.split('.')
    device_name = address_list[0]
    if device_name in DEVICES:
        element = DEVICES[device_name]
    else:
        # This should not be used on autolab closing to avoid access violation due to config opening
        element = get_device(device_name)

    for i, address_part in enumerate(address_list[1: ]):
        address_part = address_part.replace(' ', '')
        if hasattr(element, address_part):
            element = getattr(element, address_part)
        else:
            raise AttributeError(
                f"'{address_part}' not found in '{'.'.join(address_list[: i+1])}'.")
    return element


def get_final_device_config(device_name: str, **kwargs) -> dict:
    ''' Returns a valid device config from configuration file overwritten by kwargs '''
    assert device_name in list_all_devices_configs(), f"Device name '{device_name}' not found in devices_config.ini"

    # Load config object
    device_config = dict(get_device_config(device_name))

    # Overwrite config with provided configuration in kwargs
    for key, value in kwargs.items():
        device_config[key] = value

    # And the argument connection has to be provided
    assert 'driver' in device_config, f"Missing driver name for device '{device_name}'"

    if device_config['driver'] == 'autolab_server':
        device_config['connection'] = 'USELESS_ENTRY'

    assert 'connection' in device_config, f"Missing connection type for device '{device_name}'"

    return device_config


def get_device(device_name: str, **kwargs) -> Device:
    ''' Returns the Device associated to device_name. Load it if not already done.'''
    device_config = get_final_device_config(device_name, **kwargs)

    if device_name in list_loaded_devices():
        assert device_config == DEVICES[device_name].device_config, 'You cannot change the configuration of an existing Device. Close it first & retry, or remove the provided configuration.'

    else:
        instance = get_driver(
            device_config['driver'], device_config['connection'],
            **{k: v for k, v in device_config.items() if k not in [
                'driver', 'connection']})
        DEVICES[device_name] = Device(device_name, instance, device_config)
        update_allowed_dict()

    return DEVICES[device_name]


# =============================================================================
# DEVICES LIST HELP
# =============================================================================

def list_loaded_devices() -> List[str]:
    ''' Returns the list of loaded device names '''
    return list(DEVICES)


def list_devices() -> List[str]:
    ''' Returns the list of all configured device names '''
    return list_all_devices_configs()


def get_devices_status() -> dict:
    ''' Returns a dict of devices as keys and True/False values if device is loaded'''

    devices_names = list_devices()
    loaded_devices_names = list_loaded_devices()

    return {k: (k in loaded_devices_names) for k in devices_names}


# =============================================================================
# CLOSE DEVICES
# =============================================================================

def close(device: Union[str, Device] = "all"):
    """ Close a device by providing its name or its instance. Use 'all' to close all openned devices. """

    if str(device) == "all":
        for device_name in list_loaded_devices():
            try:
                DEVICES[device_name].close()
            except Exception:
                print(f"Warning: device '{device_name}' has not been closed properly")

    elif isinstance(device, Device):
        if device.name in DEVICES:
            device.close()
        else:
            print(f"No device '{device.name}' in {list_loaded_devices()}")

    elif isinstance(device, str):
        if device in DEVICES:
            DEVICES[device].close()
        else:
            print(f"No device '{device}' in {list_loaded_devices()}")

    else:
        var_type = str(type(device)).split("'")[1]
        print(f"Warning, {device} with type '{var_type}' is not a reconized device")



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 28 14:48:14 2017

@author: Quentin Chateiller
quentin.chateiller@c2n.upsaclay.fr


"""
#from threading import Event, Thread

#started = Event()

#def gui() :
#
#    if started.is_set():
#        print("Gui already running")
#    else :
#        t=AppThread()
#        t.start()


def gui():
    """ Open the Autolab GUI """
    _start('gui')


def plotter(var = None):
    """ Open the Autolab Plotter with optional variable or dataframe """
    _start('plotter', var=var)


def monitor(var, **kwargs):
    """ Open the Autolab Monitor for variable var. Optional arguments are
        mean: bool, min: bool, max: bool, delay: float, length: float """
    _start('monitor', var=var, **kwargs)


def slider(var):
    """ Open a slider for variable var """
    _start('slider', var=var)


def add_device(name: str = ''):
    """ Open the utility to add a device or modify the given device name """
    _start('add_device', var=name)


def about():
    """ Open the about window """
    _start('about')


def variables_menu():
    """ Open the variables menu """
    _start('variables_menu')


def preferences():
    """ Open the preferences window """
    _start('preferences')


def driver_installer():
    """ Open the driver installer window """
    _start('driver_installer')


def _start(gui: str, **kwargs):
    """ Open the Autolab GUI if gui='gui', the Plotter if gui='plotter'
    or the Monitor if gui='monitor' """
    import os
    from .theme import get_theme, create_stylesheet
    from ..config import get_GUI_config
    GUI_config = get_GUI_config()
    if GUI_config['qt_api'] != 'default':
        os.environ['QT_API'] = str(GUI_config['qt_api'])
    try:
        import pyqtgraph as pg
        import pyqtgraph.exporters  # Needed for pg.exporters.ImageExporter
        from qtpy import QtWidgets
    except ModuleNotFoundError as e:
        print(f"""Can't use GUI, package(s) missing: {e}
Need to install pyqtgraph, qtpy, and one of PyQt5, PySide2, PyQt6 or PySide6.

Using pip:
pip install pyqtgraph
pip install qtpy
pip install PyQt5
pip install PySide2
pip install PyQt6
pip install PySide6

Using anaconda:
conda install pyqtgraph
conda install qtpy
conda install pyqt
conda install -c conda-forge pyside2
no PyQt6 anaconda version available
conda install -c conda-forge pyside6
""")
    else:
        app = QtWidgets.QApplication.instance()
        if app is None:
            app = QtWidgets.QApplication([])

        background = GUI_config['image_background']
        foreground = GUI_config['image_foreground']

        theme_name = GUI_config['theme']
        if theme_name != 'default':
            sheet = get_theme(theme_name)
            if sheet:
                stylesheet = create_stylesheet(sheet)
                app.setStyleSheet(stylesheet)
                if theme_name == 'dark':
                    background = sheet['secondary_color']
                    foreground = sheet['text_color']
            else:
                print(f"Theme '{theme_name}' doesn't exists! Theme must be 'default' or 'dark'. Switched to 'default'")

        pg.setConfigOptions(background=background, foreground=foreground)
        pg.setConfigOption('imageAxisOrder', 'row-major')

        if GUI_config['font_size'] != 'default':
            font = app.font()
            font.setPointSize(int(float(GUI_config['font_size'])))
            app.setFont(font)

        var = kwargs.pop('var', None)

        if gui == 'gui':
            from .controlcenter.main import ControlCenter
            gui = ControlCenter()
            gui.initialize()
            gui.show()
        elif gui == 'plotter':
            from .GUI_instances import openPlotter
            openPlotter(variable=var)
        elif gui == 'monitor':
            from .GUI_instances import openMonitor
            openMonitor(var, **kwargs)
        elif gui == 'slider':
            from .GUI_instances import openSlider
            openSlider(var)
        elif gui == 'add_device':
            from .GUI_instances import openAddDevice
            openAddDevice(name=var)
        elif gui == 'about':
            from .GUI_instances import openAbout
            openAbout()
        elif gui == 'variables_menu':
            from .GUI_instances import openVariablesMenu
            openVariablesMenu()
        elif gui == 'preferences':
            from .GUI_instances import openPreferences
            openPreferences()
        elif gui == 'driver_installer':
            from .GUI_instances import openDriverInstaller
            openDriverInstaller()
        else:
            raise ValueError("gui accept either 'main', 'plotter', 'monitor'," \
                             "'slider, add_device', 'about' or 'variables_menu'" \
                             f". Given {gui}")
        app.exec()


#class AppThread(Thread):
#
#    def __init__(self):
#
#        Thread.__init__(self)
#
#    def run(self):
#
#        started.set()
#
#        _run()
#
#        started.clear()
#



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 29 18:12:24 2019

@author: qchat
"""

import os
from typing import List

import numpy as np
import pandas as pd
import pyqtgraph as pg
import pyqtgraph.exporters  # Needed for pg.exporters.ImageExporter
from qtpy import QtWidgets, QtCore

from .display import DisplayValues
from ..GUI_instances import openPlotter
from ..GUI_utilities import (get_font_size, setLineEditBackground,
                             pyqtgraph_fig_ax, pyqtgraph_image)
from ..GUI_slider import Slider
from ..icons import icons
from ...variables import Variable


if hasattr(pd.errors, 'UndefinedVariableError'):
    UndefinedVariableError = pd.errors.UndefinedVariableError
elif hasattr(pd.core.computation.ops, 'UndefinedVariableError'): # pd 1.1.5
    UndefinedVariableError = pd.core.computation.ops.UndefinedVariableError
else:
    UndefinedVariableError = Exception


class FigureManager:
    """ Manage the figure of the scanner """

    def __init__(self, gui: QtWidgets.QMainWindow):

        self.gui = gui
        self.curves = []
        self.filter_condition = []

        self._font_size = get_font_size()

        # Configure and initialize the figure in the GUI
        self.fig, self.ax = pyqtgraph_fig_ax()
        self.gui.graph.addWidget(self.fig)
        self.figMap, widget = pyqtgraph_image()
        self.gui.graph.addWidget(widget)
        self.figMap.hide()

        self.gui.variable_x_comboBox.activated.connect(
            self.axisChanged)
        self.gui.variable_x2_comboBox.activated.connect(
            self.axisChanged)
        self.gui.variable_y_comboBox.activated.connect(
            self.axisChanged)

        pgv = pg.__version__.split('.')
        if int(pgv[0]) == 0 and int(pgv[1]) < 12:
            self.gui.variable_x2_checkBox.setEnabled(False)
            self.gui.variable_x2_checkBox.setToolTip(
                "Can't use 2D plot for scan, need pyqtgraph >= 0.13.2")
            self.gui.setStatus(
                "Can't use 2D plot for scan, need pyqtgraph >= 0.13.2",
                10000, False)
        else:
            self.fig.activate_img()
            self.gui.variable_x2_checkBox.stateChanged.connect(self.reloadData)

        # Number of traces
        self.nbtraces = 5
        self.gui.nbTraces_lineEdit.setText(f'{self.nbtraces:g}')
        self.gui.nbTraces_lineEdit.returnPressed.connect(self.nbTracesChanged)
        self.gui.nbTraces_lineEdit.textEdited.connect(
            lambda: setLineEditBackground(
                self.gui.nbTraces_lineEdit, 'edited', self._font_size))
        setLineEditBackground(
            self.gui.nbTraces_lineEdit, 'synced', self._font_size)

        # Window to show scan data
        self.gui.displayScanData_pushButton.setIcon(icons['DataFrame'])
        self.gui.displayScanData_pushButton.clicked.connect(
            self.displayScanDataButtonClicked)
        self.gui.displayScanData_pushButton.hide()
        self.displayScan = DisplayValues("Scan", size=(500, 300))
        self.displayScan.setWindowIcon(icons['DataFrame'])

        self.gui.sendScanData_pushButton.setIcon(icons['plotter'])
        self.gui.sendScanData_pushButton.clicked.connect(
            self.sendScanDataButtonClicked)
        self.gui.sendScanData_pushButton.hide()

        # comboBox with scan id
        self.gui.data_comboBox.activated.connect(self.data_comboBoxClicked)
        self.gui.data_comboBox.hide()

        # Combobo to select the recipe to plot
        self.gui.scan_recipe_comboBox.activated.connect(
            self.scan_recipe_comboBoxCurrentChanged)
        self.gui.scan_recipe_comboBox.hide()

        # Combobox to select datafram to plot
        self.gui.dataframe_comboBox.activated.connect(
            self.dataframe_comboBoxCurrentChanged)
        self.gui.dataframe_comboBox.addItem("Scan")
        self.gui.dataframe_comboBox.hide()

        # Filter widgets
        self.gui.scrollArea_filter.hide()
        self.gui.checkBoxFilter.stateChanged.connect(self.checkBoxFilterChanged)
        self.gui.addFilterPushButton.clicked.connect(
            lambda: self.addFilterClicked('standard'))
        self.gui.addSliderFilterPushButton.clicked.connect(
            lambda: self.addFilterClicked('slider'))
        self.gui.addCustomFilterPushButton.clicked.connect(
            lambda: self.addFilterClicked('custom'))
        self.gui.splitterGraph.setSizes([9000, 1000])  # fixe wrong proportion

    def refresh_filters(self):
        """ Apply filters to data """
        self.filter_condition.clear()

        if self.gui.checkBoxFilter.isChecked():
            # OPTIMIZE: Should never based code on widget/layout position. Can instead create dict or other with all information about filter widgets
            for i in range(self.gui.layoutFilter.count()):
                layout = self.gui.layoutFilter.itemAt(i).widget().layout()

                if layout.count() == 5:
                    enable = bool(layout.itemAt(0).widget().isChecked())
                    variableComboBox = layout.itemAt(1).widget()
                    self.refresh_filter_combobox(variableComboBox)
                    name = variableComboBox.currentText()
                    condition_raw = layout.itemAt(2).widget().currentText()
                    valueWidget = layout.itemAt(3).widget()
                    if isinstance(valueWidget, Slider):
                        value = float(valueWidget.valueWidget.text())  # for custom slider
                        setLineEditBackground(
                            valueWidget.valueWidget, 'synced', self._font_size)
                    else:
                        try:
                            value = float(valueWidget.text())  # for editline
                        except:
                            continue
                        setLineEditBackground(
                            valueWidget, 'synced', self._font_size)

                    convert_condition = {
                        '==': np.equal, '!=': np.not_equal,
                        '<': np.less, '<=': np.less_equal,
                        '>=': np.greater_equal, '>': np.greater
                        }
                    condition = convert_condition[condition_raw]

                    filter_i = {'enable': enable, 'condition': condition,
                                'name': name, 'value': value}

                elif layout.count() == 3:
                    enable = bool(layout.itemAt(0).widget().isChecked())
                    customConditionWidget = layout.itemAt(1).widget()
                    condition_txt = customConditionWidget.text()
                    try:
                        pd.eval(condition_txt)  # syntax check
                    except UndefinedVariableError:
                        pass
                    except Exception:
                        continue
                    setLineEditBackground(
                        customConditionWidget, 'synced', self._font_size)

                    filter_i = {'enable': enable, 'condition': condition_txt,
                                'name': None, 'value': None}

                self.filter_condition.append(filter_i)

        self.reloadData()

    def refresh_filter_combobox(self, comboBox):
        items = []
        for scanset in self.gui.dataManager.datasets:
            for dataset in scanset.values():
                for key in dataset.data.columns:
                    if key not in items:
                        items.append(key)

        existing_items = [comboBox.itemText(i) for i in range(comboBox.count())]
        if items != existing_items:
            comboBox.clear()
            comboBox.addItems(items)

    def addFilterClicked(self, filter_type):
        """ Add filter condition """
        conditionWidget = QtWidgets.QFrame()
        conditionWidget.setFrameShape(QtWidgets.QFrame.StyledPanel)
        conditionLayout = QtWidgets.QHBoxLayout(conditionWidget)

        filterCheckBox = QtWidgets.QCheckBox()
        filterCheckBox.setToolTip('Toggle filter')
        filterCheckBox.setCheckState(QtCore.Qt.Checked)
        filterCheckBox.stateChanged.connect(self.refresh_filters)
        conditionLayout.addWidget(filterCheckBox)

        if filter_type in ('standard', 'slider'):
            variableComboBox = QtWidgets.QComboBox()

            self.refresh_filter_combobox(variableComboBox)
            variableComboBox.activated.connect(self.refresh_filters)
            filterCheckBox.stateChanged.connect(
                lambda: self.refresh_filter_combobox(variableComboBox))
            conditionLayout.addWidget(variableComboBox)

            filterComboBox = QtWidgets.QComboBox()
            items = ['==', '!=', '<', '<=', '>=', '>']
            filterComboBox.addItems(items)
            filterComboBox.activated.connect(self.refresh_filters)
            conditionLayout.addWidget(filterComboBox)

            if filter_type == 'standard':
                valueWidget = QtWidgets.QLineEdit()
                valueWidget.setText('1')
                valueWidget.returnPressed.connect(self.refresh_filters)
                valueWidget.textEdited.connect(lambda: setLineEditBackground(
                    valueWidget, 'edited', self._font_size))
                setLineEditBackground(valueWidget, 'synced', self._font_size)
                conditionLayout.addWidget(valueWidget)
            elif filter_type == 'slider':
                var = Variable('temp', 1.)
                valueWidget = Slider(var)
                valueWidget.minWidget.setText('1.0')
                valueWidget.minWidgetValueChanged()
                valueWidget.changed.connect(self.refresh_filters)
                valueWidget.setSizePolicy(QtWidgets.QSizePolicy.Preferred,
                                          QtWidgets.QSizePolicy.Fixed)
                conditionLayout.addWidget(valueWidget)

        elif filter_type == 'custom':
            customConditionWidget = QtWidgets.QLineEdit()
            customConditionWidget.setToolTip(
                "Filter condition can be 'id == 1' '1 <= amplitude <= 2' 'id in (1, 2)'")
            customConditionWidget.setText('id == 1')
            customConditionWidget.returnPressed.connect(self.refresh_filters)
            customConditionWidget.textEdited.connect(
                lambda: setLineEditBackground(
                    customConditionWidget, 'edited', self._font_size))
            setLineEditBackground(customConditionWidget, 'synced', self._font_size)
            conditionLayout.addWidget(customConditionWidget)

        removePushButton = QtWidgets.QPushButton()
        removePushButton.setIcon(icons['remove'])
        removePushButton.clicked.connect(
            lambda: self.remove_filter(conditionWidget))
        conditionLayout.addWidget(removePushButton)

        self.gui.layoutFilter.addWidget(conditionWidget)
        self.refresh_filters()

    def remove_filter(self, widget):
        """ Remove filter condition """
        for j in reversed(range(widget.layout().count())):
            widget.layout().itemAt(j).widget().setParent(None)
        widget.setParent(None)
        self.refresh_filters()

    def checkBoxFilterChanged(self):
        """ Show/hide filters frame and refresh filters """
        if self.gui.checkBoxFilter.isChecked():
            if not self.gui.scrollArea_filter.isVisible():
                self.gui.scrollArea_filter.show()
        else:
            if self.gui.scrollArea_filter.isVisible():
                self.gui.scrollArea_filter.hide()

        self.refresh_filters()

    def data_comboBoxClicked(self):
        """ This function select a dataset """
        if len(self.gui.dataManager.datasets) != 0:
            self.gui.data_comboBox.show()
            scanset = self.gui.dataManager.getLastSelectedDataset()
            index = self.gui.scan_recipe_comboBox.currentIndex()

            result_names = list(scanset)
            items = [self.gui.scan_recipe_comboBox.itemText(i) for i in range(
                self.gui.scan_recipe_comboBox.count())]

            if items != result_names:
                self.gui.scan_recipe_comboBox.clear()
                self.gui.scan_recipe_comboBox.addItems(result_names)
                if (index + 1) > len(result_names) or index == -1: index = 0
                self.gui.scan_recipe_comboBox.setCurrentIndex(index)

            if self.gui.scan_recipe_comboBox.count() > 1:
                self.gui.scan_recipe_comboBox.show()
            else:
                self.gui.scan_recipe_comboBox.hide()

            self.dataframe_comboBoxCurrentChanged()
        else:
            self.gui.data_comboBox.hide()

        # Change save button text to inform on scan that will be saved
        self.gui.save_pushButton.setText(
            'Save '+self.gui.data_comboBox.currentText().lower())

    def scan_recipe_comboBoxCurrentChanged(self):
        self.dataframe_comboBoxCurrentChanged()

    def dataframe_comboBoxCurrentChanged(self):

        self.updateDataframe_comboBox()

        self.gui.dataManager.updateDisplayableResults()
        self.reloadData()

        data_name = self.gui.dataframe_comboBox.currentText()

        if data_name == "Scan" or self.fig.isHidden():
            self.gui.variable_x2_checkBox.show()
        else:
            self.gui.variable_x2_checkBox.hide()

    def updateDataframe_comboBox(self):
        # Executed each time the queue is read
        index = self.gui.dataframe_comboBox.currentIndex()
        recipe_name = self.gui.scan_recipe_comboBox.currentText()
        scanset = self.gui.dataManager.getLastSelectedDataset()

        if scanset is None or recipe_name not in scanset: return None

        dataset = scanset[recipe_name]

        result_names = ["Scan"] + [
            i for i, val in dataset.data_arrays.items() if not isinstance(
                val[0], (str, tuple))]  # Remove this condition if want to plot string or tuple: Tuple[List[str], int]

        items = [self.gui.dataframe_comboBox.itemText(i) for i in range(
            self.gui.dataframe_comboBox.count())]

        if result_names != items:  # only refresh if change labels, to avoid gui refresh that prevent user to click on combobox
            self.gui.dataframe_comboBox.clear()
            self.gui.dataframe_comboBox.addItems(result_names)
            if (index + 1) > len(result_names): index = 0
            self.gui.dataframe_comboBox.setCurrentIndex(index)

        if len(result_names) == 1:
            self.gui.dataframe_comboBox.hide()
        else:
            self.gui.dataframe_comboBox.show()

    # AXE LABEL
    ###########################################################################

    def setLabel(self, axe: str, value: str):
        """ Changes the label of the given axis """
        axes = {'x':'bottom', 'y':'left'}
        if value == '': value = ' '
        self.ax.setLabel(axes[axe], value, **{'color': pg.getConfigOption("foreground"),
                                              'font-size': '12pt'})

    # PLOT DATA
    ###########################################################################

    def clearData(self):
        """ Removes any plotted curves """
        for curve in self.curves:
            try: self.ax.removeItem(curve)  # try because curve=None if close before end of scan
            except: pass
        self.curves = []
        self.figMap.clear()

        if self.fig.img_active:
            if self.fig.img.isVisible():
                self.fig.img.hide() # OPTIMIZE: would be better to erase data

    def reloadData(self):
        ''' Removes any plotted curves and reload all required
        curves from data available in the data manager'''
        # Remove all curves
        self.clearData()

        # Get current displayed result
        data_name = self.gui.dataframe_comboBox.currentText()
        variable_x = self.gui.variable_x_comboBox.currentText()
        variable_x2 = self.gui.variable_x2_comboBox.currentText()
        variable_y = self.gui.variable_y_comboBox.currentText()
        data_id = int(self.gui.data_comboBox.currentIndex()) + 1
        data_len = len(self.gui.dataManager.datasets)
        selectedData = data_len - data_id

        if variable_x is None:
            return None

        # Label update
        self.setLabel('x', variable_x)
        self.setLabel('y', variable_y)

        self.displayed_as_image = self.gui.variable_x2_checkBox.isChecked()

        if data_name == "Scan" and not self.displayed_as_image:
            nbtraces_temp  = self.nbtraces
            if not self.gui.nbTraces_lineEdit.isVisible():
                self.gui.nbTraces_lineEdit.show()
                self.gui.graph_nbTracesLabel.show()
        else:
            # decided to only show the last scan data for dataframes and scan displayed as image
            nbtraces_temp = 1
            if self.gui.nbTraces_lineEdit.isVisible():
                self.gui.nbTraces_lineEdit.hide()
                self.gui.graph_nbTracesLabel.hide()

        # Load the last results data
        if self.displayed_as_image:
            var_to_display = [variable_x, variable_x2, variable_y]
        else:
            var_to_display = [variable_x, variable_y]

        can_filter = var_to_display not in (['', ''], ['', '', ''])  # Allows to differentiate images from scan or arrays. Works only because on dataframe_comboBoxCurrentChanged, updateDisplayableResults is called
        filter_condition = self.filter_condition if (
            self.gui.checkBoxFilter.isChecked() and can_filter) else []
        data: List[pd.DataFrame] = self.gui.dataManager.getData(
            nbtraces_temp, var_to_display,
            selectedData=selectedData, data_name=data_name,
            filter_condition=filter_condition)

        # Plot data
        if data is not None:
            true_nbtraces = max(nbtraces_temp, len(data))  # not good but avoid error

            if len(data) != 0:
                # update displayScan
                self.refreshDisplayScanData()
                if not self.gui.displayScanData_pushButton.isVisible():
                    self.gui.displayScanData_pushButton.show()
                    self.gui.sendScanData_pushButton.show()

                for temp_data in data:
                    if temp_data is not None: break
                else: return None

            # If plot scan as image
            if data_name == "Scan" and self.displayed_as_image:

                if not self.gui.frameAxis.isVisible():
                    self.gui.frameAxis.show()

                if not self.gui.variable_x2_comboBox.isVisible():
                    self.gui.variable_x2_comboBox.show()
                    self.gui.label_scan_2D.show()
                    self.gui.label_y_axis.setText('Z axis')

                if not self.fig.isVisible():
                    self.figMap.hide()
                    self.fig.show()

                if not self.fig.colorbar.isVisible():
                    self.fig.colorbar.show()

                self.setLabel('x', variable_x)
                self.setLabel('y', variable_x2)

                if variable_x == variable_x2:
                    return None

                # Data
                if len(data) == 0:
                    return None

                subdata: pd.DataFrame = data[-1]  # Only plot last scan

                if subdata is None:
                    return None

                subdata = subdata.astype(float)

                try:
                    pivot_table = subdata.pivot(
                        index=variable_x, columns=variable_x2, values=variable_y)
                except ValueError:  # if more than 2 parameters
                    return None

                # Extract data for plotting
                x = np.array(pivot_table.columns)
                y = np.array(pivot_table.index)
                z = np.array(pivot_table)

                if 0 in (len(x), len(y), len(z)):
                    return None

                self.fig.update_img(x, y, z)

                if not self.fig.img.isVisible():
                    self.fig.img.show()

                return None

            # If plot scan or array
            if self.gui.variable_x2_comboBox.isVisible():
                self.gui.variable_x2_comboBox.hide()
                self.gui.label_scan_2D.hide()
                self.gui.label_y_axis.setText('Y axis')

            if self.fig.img_active:
                if self.fig.colorbar.isVisible():
                    self.fig.colorbar.hide()

                if self.fig.img.isVisible():
                    self.fig.img.hide()

            if len(data) != 0 and isinstance(data[0], np.ndarray):  # to avoid errors
                images = np.empty((len(data), *temp_data.shape))

            for i in range(len(data)):
                # Data
                subdata: pd.DataFrame = data[i]

                if subdata is None: continue

                if isinstance(subdata, str):  # Could think of someway to show text. Currently removed it from dataset directly
                    print("Warning: Can't display text")
                    continue
                if isinstance(subdata, tuple):
                    print("Warning: Can't display tuple")
                    continue

                subdata = subdata.astype(float)

                if isinstance(subdata, np.ndarray):  # is image
                    if self.fig.isVisible():
                        self.fig.hide()
                        self.figMap.show()
                    if self.gui.frameAxis.isVisible():
                        self.gui.frameAxis.hide()
                    # OPTIMIZE: should be able to plot images of different shapes
                    if subdata.shape == temp_data.shape:
                        images[i] = subdata
                        has_bad_shape = False
                    else:
                        has_bad_shape = True
                    if i == len(data)-1:
                        if images.ndim == 3:
                            if self.figMap.imageItem.axisOrder == 'col-major':
                                x, y = (0, 1)
                            else:
                                x, y = (1, 0)
                            axes = {'t': 0, 'x': x+1, 'y': y+1, 'c': None}  # to avoid a special case in pg that incorrectly assumes the axis
                        else:
                            axes = None
                        if has_bad_shape:
                            images_plot = np.array([subdata])
                            print(f"Warning only plot last {data_name}." \
                                  " Images should have the same shape." \
                                  f" Given {subdata.shape} & {temp_data.shape}")
                        else:
                            images_plot = images
                        try:
                            self.figMap.setImage(images_plot, axes=axes)  # Defined which axe is major using pg.setConfigOption('imageAxisOrder', 'row-major') in gui start-up so no need to .T data
                        except Exception as e:
                            print(f"Warning can't plot {data_name}: {e}")
                            continue
                        self.figMap.setCurrentIndex(len(self.figMap.tVals))

                else: # not an image (is pd.DataFrame)
                    if not self.fig.isVisible():
                        self.fig.show()
                        self.figMap.hide()
                    if not self.gui.frameAxis.isVisible():
                        self.gui.frameAxis.show()

                    try:  # If empty dataframe can't find variables
                        x = subdata.loc[:, variable_x]
                        y = subdata.loc[:, variable_y]
                    except KeyError:
                        continue

                    if isinstance(x, pd.DataFrame):
                        print('Warning: At least two variables have the same name.' \
                              f" Data plotted is incorrect for {variable_x}!")
                    if isinstance(y, pd.DataFrame):
                        print('Warning: At least two variables have the same name.' \
                              f" Data plotted is incorrect for {variable_y}!")
                        y = y.iloc[:, 0]

                    if i == (len(data) - 1):
                        color = 'r'
                        alpha = 1
                    else:
                        color = pg.getConfigOption("foreground")
                        alpha = (true_nbtraces - (len(data) - 1 - i)) / true_nbtraces

                    # TODO: no information about dataset nor scanset in this method! See what could be done
                    # if scanset.color != 'default':
                    #     color = scanset.color

                    # Plot
                    # careful, now that can filter data, need .values to avoid pyqtgraph bug
                    # pyqtgraph 0.11.1 raise hover error if plot deleted
                    curve = self.ax.plot(x.values, y.values, symbol='x',
                                         symbolPen=color, symbolSize=10,
                                         pen=color, symbolBrush=color)
                    curve.setAlpha(alpha, False)
                    self.curves.append(curve)

    def axisChanged(self, index):
        """ Called when the displayed result has been changed
        in the combo box. It proceeds to the change. """
        if (self.gui.variable_x_comboBox.currentIndex() != -1
                and self.gui.variable_y_comboBox.currentIndex() != -1):
            self.reloadData()
        else:
            self.clearData()

    # TRACES
    ###########################################################################

    def nbTracesChanged(self):
        """ Called when the number of traces displayed has been changed
        in the GUI. It proceeds to the change and update the plot. """
        value = self.gui.nbTraces_lineEdit.text()
        check = False
        try:
            value = int(float(value))
            assert value > 0
            self.nbtraces = value
            check = True
        except:
            pass

        self.gui.nbTraces_lineEdit.setText(f'{self.nbtraces:g}')
        setLineEditBackground(self.gui.nbTraces_lineEdit, 'synced', self._font_size)

        if check is True and self.gui.variable_y_comboBox.currentIndex() != -1:
            self.reloadData()

    # Show data
    ###########################################################################
    def refreshDisplayScanData(self):
        recipe_name = self.gui.scan_recipe_comboBox.currentText()
        scanset = self.gui.dataManager.getLastSelectedDataset()
        if scanset is not None and recipe_name in scanset:
            name = f"Scan{self.gui.data_comboBox.currentIndex()+1}"
            if self.gui.scan_recipe_comboBox.count() > 1:
                name += f", {recipe_name}"
            self.displayScan.setWindowTitle(name)
            self.displayScan.refresh(scanset[recipe_name].data)

    def displayScanDataButtonClicked(self):
        """ Opens a window showing the scan data for the displayed scan id """
        self.refreshDisplayScanData()
        self.displayScan.show()

    def sendScanDataButtonClicked(self):
        """ Sends the displayed scan data to plotter """
        recipe_name = self.gui.scan_recipe_comboBox.currentText()
        scanset = self.gui.dataManager.getLastSelectedDataset()
        if scanset is not None and recipe_name in scanset:
            data = scanset[recipe_name].data
            scan_name = self.gui.data_comboBox.currentText()
            data.name = f"{scan_name}_{recipe_name}"
            openPlotter(variable=scanset[recipe_name].data, has_parent=True)

    # SAVE FIGURE
    ###########################################################################

    def save(self, filename: str):
        """ Saves the figure with the provided filename """
        raw_name, extension = os.path.splitext(filename)
        new_filename = raw_name + ".png"
        exporter = pg.exporters.ImageExporter(self.ax)
        exporter.export(new_filename)

    def close(self):
        """ Called by scanner on closing """
        self.displayScan.close()
        self.fig.deleteLater()  # prevent crash without traceback when reopenning scanner multiple times
        self.figMap.deleteLater()



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 29 18:08:45 2019

@author: qchat
"""

import os
import time
import math as m
import threading
from collections import OrderedDict
from queue import Queue
from itertools import product

import numpy as np
from qtpy import QtCore, QtWidgets

from ..GUI_utilities import qt_object_exists, MyInputDialog, MyFileDialog
from ..GUI_instances import instances
from ...paths import PATHS
from ...variables import eval_variable, set_variable, has_eval
from ...utilities import create_array


class ScanManager:
    """ Manage a scan using a dedicated thread """

    def __init__(self, gui: QtWidgets.QMainWindow):
        self.gui = gui

        # Start / Stop button configuration
        self.gui.start_pushButton.clicked.connect(self.startButtonClicked)
        self.gui.stop_pushButton.clicked.connect(self.stopButtonClicked)
        self.gui.stop_pushButton.setEnabled(False)

        # Pause / Resume button configuration
        self.gui.pause_pushButton.clicked.connect(self.pauseButtonClicked)
        self.gui.pause_pushButton.setEnabled(False)

        # Progress bar configuration
        self.gui.progressBar.setMinimum(0)
        self.gui.progressBar.setValue(0)

        # Thread
        self.thread = None
        self.main_dialog = None

    # START AND STOP
    #############################################################################

    def startButtonClicked(self):
        """ Called when the start button is pressed.
        Do the expected action """
        if not self.isStarted():
            self.start()

    def stopButtonClicked(self):
        """ Called when the stop button is pressed.
        Do the expected action """
        if self.isStarted():
            self.stop()

    def isStarted(self):
        """ Returns True or False whether the scan is currently running or not """
        return self.thread is not None

    def start(self):
        """ Starts a scan """
        try:
            self.gui.configManager.checkConfig()  #  raise error if config not valid
            config = self.gui.configManager.config
        except Exception as e:
            self.gui.setStatus(f'ERROR The scan cannot start with the current configuration: {str(e)}', 10000, False)
            return None

        # Should not be possible
        if self.thread is not None:
            self.gui.setStatus('ERROR: A scan thread already exists!', 10000, False)
            try: self.thread.finished()
            except: pass
            self.thread = None
            return None

        # Only if current config is valid to start a scan

        # Pause monitors if option selected in monitors
        for var_id in set([id(step['element'])
                       for recipe in config.values()
                       for step in recipe['recipe']+recipe['parameter']]):
            if var_id in instances['monitors']:
                monitor = instances['monitors'][var_id]
                if (monitor.pause_on_scan
                        and not monitor.monitorManager.isPaused()):
                    monitor.pauseButtonClicked()

        # Prepare a new dataset in the datacenter
        self.gui.dataManager.newDataset(config)

        # put dataset id onto the combobox and associate data to it
        dataSet_id = len(self.gui.dataManager.datasets)
        self.gui.data_comboBox.addItem(f'scan{dataSet_id}')
        self.gui.data_comboBox.setCurrentIndex(int(dataSet_id)-1)  # trigger the currentIndexChanged event but don't trigger activated

        # Start a new thread
        ## Opening
        self.thread = ScanThread(self.gui.dataManager.queue, config)
        ## Signal connections
        self.thread.errorSignal.connect(self.error)
        self.thread.userSignal.connect(self.handler_user_input)

        self.thread.startParameterSignal.connect(
            lambda recipe_name, param_name: self.setParameterProcessingState(
                recipe_name, param_name, 'started'))
        self.thread.finishParameterSignal.connect(
            lambda recipe_name, param_name: self.setParameterProcessingState(
                recipe_name, param_name, 'finished'))
        self.thread.parameterCompletedSignal.connect(
            lambda recipe_name, param_name: self.resetParameterProcessingState(
                recipe_name, param_name))

        self.thread.startStepSignal.connect(
            lambda recipe_name, stepName: self.setStepProcessingState(
                recipe_name, stepName, 'started'))
        self.thread.finishStepSignal.connect(
            lambda recipe_name, stepName: self.setStepProcessingState(
                recipe_name, stepName, 'finished'))
        self.thread.recipeCompletedSignal.connect(
            lambda recipe_name: self.resetStepsProcessingState(recipe_name))
        self.thread.scanCompletedSignal.connect(self.scanCompleted)

        self.thread.finished.connect(self.finished)

        # Starting
        self.thread.start()

        # Start data center timer
        self.gui.dataManager.timer.start()

        # Update gui
        self.gui.start_pushButton.setEnabled(False)
        self.gui.stop_pushButton.setEnabled(True)
        self.gui.pause_pushButton.setEnabled(True)
        self.gui.clear_pushButton.setEnabled(False)
        self.gui.progressBar.setValue(0)
        self.gui.progressBar.setStyleSheet("")
        self.gui.importAction.setEnabled(False)
        self.gui.openRecentMenu.setEnabled(False)
        self.gui.undo.setEnabled(False)
        self.gui.redo.setEnabled(False)
        self.gui.setStatus('Scan started!', 5000)
        self.gui.refresh_widget(self.gui.start_pushButton)

    def handler_user_input(self, stepInfos: dict):
        unit = stepInfos['element'].unit
        name = stepInfos['name']

        if unit in ("open-file", "save-file"):

            if unit == "open-file":
                self.main_dialog = MyFileDialog(self.gui, name,
                                                QtWidgets.QFileDialog.AcceptOpen)
                self.main_dialog.show()

                if self.main_dialog.exec_() == QtWidgets.QInputDialog.Accepted:
                    filename = self.main_dialog.selectedFiles()[0]
                else:
                    filename = ''

            elif unit == "save-file":
                self.main_dialog = MyFileDialog(self.gui, name,
                                                QtWidgets.QFileDialog.AcceptSave)
                self.main_dialog.show()

                if self.main_dialog.exec_() == QtWidgets.QInputDialog.Accepted:
                    filename = self.main_dialog.selectedFiles()[0]
                else:
                    filename = ''

            if filename != '':
                path = os.path.dirname(filename)
                PATHS['last_folder'] = path

            if qt_object_exists(self.main_dialog): self.main_dialog.deleteLater()
            if self.thread is not None: self.thread.user_response = filename

        elif unit == 'user-input':
            self.main_dialog = MyInputDialog(self.gui, name)
            self.main_dialog.show()

            if self.main_dialog.exec_() == QtWidgets.QInputDialog.Accepted:
                response = self.main_dialog.textValue()
            else:
                response = ''

            if qt_object_exists(self.main_dialog): self.main_dialog.deleteLater()

            if has_eval(response):
                try:
                    response = eval_variable(response)
                except Exception as e:
                    self.thread.errorSignal.emit(e)
                    self.thread.stopFlag.set()

            if self.thread is not None: self.thread.user_response = response
        else:
            if self.thread is not None: self.thread.user_response = f"Unknown unit '{unit}'"

    def scanCompleted(self):
        if not qt_object_exists(self.gui.progressBar):
            return None

        if self.thread.stopFlag.is_set():
            self.gui.progressBar.setStyleSheet(
                "QProgressBar::chunk {background-color: red;}")
            # self.gui.setStatus('Scan stopped!', 5000)  # not good because hide error message
        else:
            self.gui.setStatus('Scan finished!', 5000)
            self.gui.progressBar.setMaximum(1)
            self.gui.progressBar.setValue(1)
            self.gui.progressBar.setStyleSheet("")

            # Start monitors if option selected in monitors
            for var_id in set([id(step['element'])
                           for recipe in self.thread.config.values()
                           for step in recipe['recipe']+recipe['parameter']]):
                if var_id in instances['monitors']:
                    monitor = instances['monitors'][var_id]
                    if (monitor.start_on_scan
                            and monitor.monitorManager.isPaused()):
                        monitor.pauseButtonClicked()

    def setStepProcessingState(self, recipe_name: str, stepName: str, state: str):
        self.gui.recipeDict[recipe_name]['recipeManager'].setStepProcessingState(stepName, state)

    def resetStepsProcessingState(self, recipe_name: str):
        self.gui.recipeDict[recipe_name]['recipeManager'].resetStepsProcessingState()

    def setParameterProcessingState(self, recipe_name: str, param_name: str, state: str):
        self.gui.recipeDict[recipe_name]['parameterManager'][param_name].setProcessingState(state)

    def resetParameterProcessingState(self, recipe_name: str, param_name: str):
        self.gui.recipeDict[recipe_name]['parameterManager'][param_name].setProcessingState('idle')

    def stop(self):
        """ Stops manually the scan """
        self.disableContinuousMode()
        self.thread.stopFlag.set()
        self.thread.user_response = 'Close'  # needed to stop scan
        self.resume()
        if self.main_dialog and qt_object_exists(self.main_dialog):
            self.main_dialog.deleteLater()
        self.thread.wait()
    # SIGNALS
    #############################################################################

    def finished(self):
        """ This function is called when the scan thread is finished.
        It restores the GUI in a ready mode, and start a new scan if in
        continuous mode """
        if not qt_object_exists(self.gui.stop_pushButton):
            return None
        self.gui.stop_pushButton.setEnabled(False)
        self.gui.pause_pushButton.setEnabled(False)
        self.gui.clear_pushButton.setEnabled(True)
        self.gui.importAction.setEnabled(True)
        self.gui.openRecentMenu.setEnabled(True)
        self.gui.configManager.updateUndoRedoButtons()
        self.gui.dataManager.timer.stop()
        self.gui.dataManager.sync() # once again to be sure we grabbed every data
        self.thread = None
        self.gui.refresh_widget(self.gui.stop_pushButton)

        if self.isContinuousModeEnabled():
            self.start()
        else:
            self.gui.start_pushButton.setEnabled(True)

    def error(self, error: Exception):
        """ Called if an error occured during the scan.
        It displays it in the status bar """
        self.gui.setStatus(f'Scan error: {error}', 10000, False)


    # CONTINUOUS MODE
    #############################################################################

    def isContinuousModeEnabled(self) -> bool:
        """ Returns True or False whether the continuous scan mode is enabled
        or not """
        return self.gui.continuous_checkBox.isChecked()

    def disableContinuousMode(self):
        """ Disables the continuous scan mode """
        self.gui.continuous_checkBox.setChecked(False)

    # PAUSE
    #############################################################################

    def pauseButtonClicked(self):
        """ Called when the Pause/Resume button is clicked.
        It does the required action """
        if self.isStarted():
            self.resume() if self.isPaused() else self.pause()

    def isPaused(self):
        """ Returns True or False whether the scan is paused or not """
        return (self.thread is not None) and (self.thread.pauseFlag.is_set() is True)

    def pause(self):
        """ Pauses the scan """
        self.thread.pauseFlag.set()
        self.gui.dataManager.timer.stop()
        self.gui.dataManager.sync() # once again to be sure we grabbed every data
        self.gui.pause_pushButton.setText('Resume')

    def resume(self):
        """ Resumes the scan """
        try:
            # One possible error is device closed while scan was paused
            self.gui.configManager.checkConfig()  #  raise error if config not valid
        except Exception as e:
            self.gui.setStatus(f'WARNING The scan cannot resume: {str(e)}', 10000, False)
            return None
        self.thread.pauseFlag.clear()
        self.gui.dataManager.timer.start()
        self.gui.pause_pushButton.setText('Pause')


class ScanThread(QtCore.QThread):
    """ This thread class is dedicated to read the variable,
        and send its data to GUI through a queue """
    # Signals
    userSignal = QtCore.Signal(dict)
    errorSignal = QtCore.Signal(object)

    startParameterSignal = QtCore.Signal(object, object)
    finishParameterSignal = QtCore.Signal(object, object)
    parameterCompletedSignal = QtCore.Signal(object, object)

    startStepSignal = QtCore.Signal(object, object)
    finishStepSignal = QtCore.Signal(object, object)
    recipeCompletedSignal = QtCore.Signal(object)
    scanCompletedSignal = QtCore.Signal()

    def __init__(self, queue: Queue, config: dict):
        super().__init__()
        self.config = config
        self.queue = queue

        self.pauseFlag = threading.Event()
        self.stopFlag = threading.Event()

        self.user_response = None

    def run(self):
        # Start the scan
        for recipe_name in self.config:
            if self.config[recipe_name]['active']: self.execRecipe(recipe_name)

        self.scanCompletedSignal.emit()

    def execRecipe(self, recipe_name: str,
                   initPoint: OrderedDict = None):
        """ Executes a recipe. initPoint is obsolete, was used to add parameters values
        and master-recipe name to a sub-recipe """

        paramValues_list = []


        for parameter in self.config[recipe_name]['parameter']:
            param_name = parameter['name']

            if 'values' in parameter:
                paramValues = parameter['values']
                try:
                    paramValues = eval_variable(paramValues)
                    paramValues = create_array(paramValues)
                except Exception as e:
                    self.errorSignal.emit(e)
                    self.stopFlag.set()
            else:
                startValue, endValue = parameter['range']
                nbpts = parameter['nbpts']
                logScale = parameter['log']

                # Creates the array of values for the parameter
                if logScale:
                    paramValues = np.logspace(m.log10(startValue), m.log10(endValue), nbpts, endpoint=True)
                else:
                    paramValues = np.linspace(startValue, endValue, nbpts, endpoint=True)

            set_variable(param_name, paramValues[0])
            paramValues_list.append(paramValues)

        ID = 0
        # iter over each parameter (do once if no parameter!)
        for i, paramValueList in enumerate(product(*paramValues_list)):

            if not self.stopFlag.is_set():

                if initPoint is None:  # OBSOLETE
                    initPoint = OrderedDict()
                    initPoint[0] = recipe_name

                initPointStep = initPoint.copy()

                try:
                    self._source_of_error = None
                    ID += 1
                    set_variable('ID', ID)

                    for parameter, paramValue in zip(
                            self.config[recipe_name]['parameter'], paramValueList):
                        self._source_of_error = parameter
                        element = parameter['element']
                        param_name = parameter['name']

                        set_variable(param_name, element.type(
                                paramValue) if element is not None else paramValue)

                        # Set the parameter value
                        self.startParameterSignal.emit(recipe_name, param_name)
                        if element is not None: element(paramValue)
                        self.finishParameterSignal.emit(recipe_name, param_name)

                        initPointStep[param_name] = paramValue

                    dataPoint = initPointStep.copy()

                    # Start the recipe
                    dataPoint = self.processStep(
                        recipe_name, dataPoint, initPointStep)
                    # Send the whole data in the queue
                    if not self.stopFlag.is_set(): self.queue.put(dataPoint)

                except Exception as e:
                    # If an error occurs, stop the scan and send an error signal
                    name = self._source_of_error['name']
                    if self._source_of_error['element'] is not None:
                        address = f"='{self._source_of_error['element'].address()}'"
                    else: address = ''

                    try:
                        from pyvisa import VisaIOError
                    except:
                        e = f"In recipe '{recipe_name}' for step '{name}': {e}"
                    else:
                        if str(e) == str(VisaIOError(-1073807339)):
                            e = f"Timeout reached for device {address}. Acquisition time may be too long. If so, you can increase timeout delay in the driver to avoid this error."
                        else:
                            e = f"In recipe '{recipe_name}' for step '{name}': {e}"

                    self.errorSignal.emit(e)
                    self.stopFlag.set()

                # Wait until the scan is no more in pause
                while self.pauseFlag.is_set():
                    time.sleep(0.1)
            else:
                break

        for parameter in self.config[recipe_name]['parameter']:
            self.parameterCompletedSignal.emit(recipe_name, parameter['name'])

    def processStep(self, recipe_name: str,
                    dataPoint: OrderedDict,
                    initPoint: OrderedDict):
        """ Executes the recipe step """
        for stepInfos in self.config[recipe_name]['recipe']:
            self._source_of_error = stepInfos

            if not self.stopFlag.is_set():
                # Process the recipe step
                result = self.processElement(recipe_name, stepInfos, initPoint)

                if result is not None:
                    dataPoint[stepInfos['name']] = result

                # Wait until the scan is no more in pause
                while self.pauseFlag.is_set():
                    time.sleep(0.1)
            else:
                break

        self.recipeCompletedSignal.emit(recipe_name)
        return dataPoint

    def processElement(self, recipe_name: str, stepInfos: dict,
                       initPoint: OrderedDict):
        """ Processes the recipe step """
        element = stepInfos['element']
        stepType = stepInfos['stepType']
        self.startStepSignal.emit(recipe_name, stepInfos['name'])
        result = None

        if stepType == 'measure':
            result = element()
            set_variable(stepInfos['name'], result)
        elif stepType == 'set':
            value = eval_variable(stepInfos['value'])
            if element.type in [bytes] and isinstance(value, str): value = value.encode()
            if element.type in [np.ndarray]: value = create_array(value)
            element(value)
        elif stepType == 'action':
            if stepInfos['value'] is not None:
                # Open dialog for open file, save file or input text
                if isinstance(stepInfos['value'], str) and stepInfos['value'] == '':
                    self.userSignal.emit(stepInfos)
                    while (not self.stopFlag.is_set()
                           and self.user_response is None):
                        time.sleep(0.1)
                    if not self.stopFlag.is_set():
                        element(self.user_response)
                    self.user_response = None
                else:
                    value = eval_variable(stepInfos['value'])
                    if element.type in [bytes] and isinstance(value, str): value = value.encode()
                    if element.type in [np.ndarray]: value = create_array(value)
                    element(value)
            else:
                element()
        elif stepType == 'recipe':  # OBSOLETE
            self.execRecipe(element, initPoint=initPoint)  # Execute a recipe in the recipe

        self.finishStepSignal.emit(recipe_name, stepInfos['name'])
        return result



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 14 18:01:42 2024

@author: Jonathan
"""


import os
import sys

from .utilities import input_wrap


def create_shortcut(ask: bool = False):
    """ Create Autolab GUI shortcut on desktop. """
    # Works on Windows with winpython and with base or env conda
    try:
        autolab_icon = os.path.join(
            os.path.dirname(__file__), 'gui',
            'icons', 'autolab-icon.ico').replace("\\", "/")
        userprofile = os.path.expanduser('~')
        desktop = os.path.join(userprofile, 'Desktop')
        link = os.path.join(desktop, 'Autolab GUI.lnk')
        python = sys.base_prefix
        python_env = sys.prefix

        if not os.path.exists(desktop):
            return None

        is_conda = os.path.exists(os.path.join(sys.prefix, 'conda-meta'))

        if is_conda:
            if python == python_env:
                if 'envs' in python:
                    python = os.path.dirname(os.path.dirname(python))
        else:  # Suppose WinPython
            python = os.path.dirname(python)

        python_script = os.path.normpath(os.path.join(python, r'Scripts/activate.bat'))

        if not os.path.exists(python_script):
            return None

        from comtypes.client import CreateObject
        from comtypes.persist import IPersistFile
        from comtypes.shelllink import ShellLink

        if ask:
            ans = input_wrap('Create Autolab GUI shortcut on desktop? [default:yes] > ')
        else:
            ans = 'yes'

        if ans.strip().lower() == 'no':
            return None

        s = CreateObject(ShellLink)
        s.SetPath('cmd.exe')
        s.SetArguments(f'/c {python_script} {python_env} && autolab gui')
        s.SetIconLocation(autolab_icon, 0)
        s.SetDescription('Start Autolab GUI')
        s.SetWorkingDirectory(userprofile)

        p = s.QueryInterface(IPersistFile)
        p.Save(link, True)

        if not os.path.exists(os.path.join(python_env, r'Scripts/autolab.exe')):
            print('Warning autolab has not been installed with pip, the shortcut will not work')

    except Exception as e:
        print(f'Cannot create Autolab shortcut: {e}')



# -*- coding: utf-8 -*-
"""
Created on Sat Aug  3 20:40:00 2024

@author: jonathan
"""

from typing import Union, Any

import pandas as pd
from qtpy import QtWidgets, QtCore

from ..devices import get_final_device_config

from ..elements import Variable as Variable_og
from ..variables import Variable

# Contains local import:
# from .monitoring.main import Monitor
# from .GUI_slider import Slider
# from .GUI_variables import VariablesMenu
# from .plotting.main import Plotter
# from .GUI_add_device import AddDeviceWindow
# from .GUI_about import AboutWindow
# Not yet or maybe never (too intertwined with mainGui) # from .scanning.main import Scanner


instances = {
    'monitors': {},
    'sliders': {},
    'variablesMenu': None,
    'plotter': None,
    'addDevice': None,
    'about': None,
    'preferences': None,
    'driverInstaller': None,
    # 'scanner': None,
}


# =============================================================================
# Monitor
# =============================================================================
def openMonitor(variable: Union[Variable, Variable_og],
                has_parent: bool = False, **kwargs):
    """ Opens the monitor associated to the variable. """
    from .monitoring.main import Monitor  # Inside to avoid circular import

    assert isinstance(variable, (Variable, Variable_og)), (
        f'Need type {Variable} or {Variable_og}, but given type is {type(variable)}')
    assert variable.readable, f"The variable {variable.address()} is not readable"

    # If the monitor is not already running, create one
    if id(variable) not in instances['monitors'].keys():
        instances['monitors'][id(variable)] = Monitor(variable, has_parent, **kwargs)
        instances['monitors'][id(variable)].show()
    # If the monitor is already running, just make as the front window
    else:
        monitor = instances['monitors'][id(variable)]
        monitor.setWindowState(
            monitor.windowState()
            & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
        monitor.activateWindow()


def clearMonitor(variable: Union[Variable, Variable_og]):
    """ Clears monitor instances reference when quitted """
    if id(variable) in list(instances['monitors']):
        instances['monitors'].pop(id(variable))


def closeMonitors():
    for monitor in list(instances['monitors'].values()):
        monitor.close()


# =============================================================================
# Slider
# =============================================================================
def openSlider(variable: Union[Variable, Variable_og],
               gui: QtWidgets.QMainWindow = None,
               item: QtWidgets.QTreeWidgetItem = None):
    """ Opend the slider associated to this variable. """
    from .GUI_slider import Slider  # Inside to avoid circular import

    assert isinstance(variable, (Variable, Variable_og)), (
        f'Need type {Variable} or {Variable_og}, but given type is {type(variable)}')
    assert variable.writable, f"The variable {variable.address()} is not writable"

    # If the slider is not already running, create one
    if id(variable) not in instances['sliders'].keys():
        instances['sliders'][id(variable)] = Slider(variable, gui=gui, item=item)
        instances['sliders'][id(variable)].show()
    # If the slider is already running, just make as the front window
    else:
        slider = instances['sliders'][id(variable)]
        slider.setWindowState(
            slider.windowState()
            & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
        slider.activateWindow()


def clearSlider(variable: Union[Variable, Variable_og]):
    """ Clears the slider instances reference when quitted """
    if id(variable) in instances['sliders'].keys():
        instances['sliders'].pop(id(variable))


def closeSliders():
    for slider in list(instances['sliders'].values()):
        slider.close()


# =============================================================================
# VariableMenu
# =============================================================================
def openVariablesMenu(has_parent: bool = False):
    """ Opens the variables menu. """
    from .GUI_variables import VariablesMenu  # Inside to avoid circular import
    if instances['variablesMenu'] is None:
        instances['variablesMenu'] = VariablesMenu(has_parent)
        instances['variablesMenu'].show()
    else:
        instances['variablesMenu'].refresh()
        instances['variablesMenu'].setWindowState(
            instances['variablesMenu'].windowState()
            & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
        instances['variablesMenu'].activateWindow()


def clearVariablesMenu():
    """ Clears the variables menu instance reference when quitted """
    instances['variablesMenu'] = None


def closeVariablesMenu():
    if instances['variablesMenu'] is not None:
        instances['variablesMenu'].close()


# =============================================================================
# Plotter
# =============================================================================
def openPlotter(variable: Union[Variable, Variable_og, pd.DataFrame, Any] = None,
                has_parent: bool = False):
    """ Opens the plotter. Can add variable. """
    from .plotting.main import Plotter  # Inside to avoid circular import
    # If the plotter is not already running, create one
    if instances['plotter'] is None:
        instances['plotter'] = Plotter(has_parent)
    # If the plotter is not active open it (keep data if closed)
    if not instances['plotter'].active:
        instances['plotter'].show()
        instances['plotter'].activateWindow()
        instances['plotter'].active = True
    # If the plotter is already running, just make as the front window
    else:
        instances['plotter'].setWindowState(
            instances['plotter'].windowState()
            & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
        instances['plotter'].activateWindow()

    if variable is not None:
            instances['plotter'].refreshPlotData(variable)


def clearPlotter():
    """ Deactivates the plotter when quitted but keep the instance in memory """
    if instances['plotter'] is not None:
        instances['plotter'].active = False  # don't want to close plotter because want to keep data


def closePlotter():
    if instances['plotter'] is not None:
        instances['plotter'].figureManager.fig.deleteLater()
        for children in instances['plotter'].findChildren(QtWidgets.QWidget):
            children.deleteLater()

        instances['plotter'].close()
        instances['plotter'] = None  # To remove plotter from memory


# =============================================================================
# AddDevice
# =============================================================================
def openAddDevice(gui: QtWidgets.QMainWindow = None, name: str = ''):
    """ Opens the add device window. """
    from .GUI_add_device import AddDeviceWindow  # Inside to avoid circular import
    # If the add device window is not already running, create one
    if instances['addDevice'] is None:
        instances['addDevice'] = AddDeviceWindow(gui)
        instances['addDevice'].show()
        instances['addDevice'].activateWindow()
    # If the add device window is already running, just make as the front window
    else:
        instances['addDevice'].setWindowState(
            instances['addDevice'].windowState()
            & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
        instances['addDevice'].activateWindow()

    # Modify existing device
    if name != '':
        try:
            conf = get_final_device_config(name)
        except Exception as e:
            instances['addDevice'].setStatus(str(e), 10000, False)
        else:
            instances['addDevice'].modify(name, conf)


def clearAddDevice():
    """ Clears the addDevice instance reference when quitted """
    instances['addDevice'] = None


def closeAddDevice():
    if instances['addDevice'] is not None:
        instances['addDevice'].close()


# =============================================================================
# About
# =============================================================================
def openAbout(gui: QtWidgets.QMainWindow = None):
    """ Opens the about window. """
    # If the about window is not already running, create one
    from .GUI_about import AboutWindow  # Inside to avoid circular import
    if instances['about'] is None:
        instances['about'] = AboutWindow(gui)
        instances['about'].show()
        instances['about'].activateWindow()
    # If the about window is already running, just make as the front window
    else:
        instances['about'].setWindowState(
            instances['about'].windowState()
            & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
        instances['about'].activateWindow()


def clearAbout():
    """ Clears the about instance reference when quitted """
    instances['about'] = None


def closeAbout():
    if instances['about'] is not None:
        instances['about'].close()


# =============================================================================
# Preferences
# =============================================================================
def openPreferences(gui: QtWidgets.QMainWindow = None):
    """ Opens the preferences window. """
    # If the about window is not already running, create one
    from .GUI_preferences import PreferencesWindow  # Inside to avoid circular import
    if instances['preferences'] is None:
        instances['preferences'] = PreferencesWindow(gui)
        instances['preferences'].show()
        instances['preferences'].activateWindow()
    # If the about window is already running, just make as the front window
    else:
        instances['preferences'].setWindowState(
            instances['preferences'].windowState()
            & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
        instances['preferences'].activateWindow()


def clearPreferences():
    """ Clears the preferences instance reference when quitted """
    instances['preferences'] = None


def closePreferences():
    if instances['preferences'] is not None:
        instances['preferences'].close()



# =============================================================================
# Driver installer
# =============================================================================
def openDriverInstaller(gui: QtWidgets.QMainWindow = None):
    """ Opens the driver installer. """
    # If the about window is not already running, create one
    from .GUI_driver_installer import DriverInstaller  # Inside to avoid circular import
    if instances['driverInstaller'] is None:
        instances['driverInstaller'] = DriverInstaller(gui)
        instances['driverInstaller'].show()
        instances['driverInstaller'].activateWindow()
    # If the about window is already running, just make as the front window
    else:
        instances['driverInstaller'].setWindowState(
            instances['driverInstaller'].windowState()
            & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
        instances['driverInstaller'].activateWindow()


def clearDriverInstaller():
    """ Clears the driver unstaller instance reference when quitted """
    instances['driverInstaller'] = None


def closeDriverInstaller():
    if instances['driverInstaller'] is not None:
        instances['driverInstaller'].close()


# =============================================================================
# Scanner
# =============================================================================
# def openScanner(gui: QtWidgets.QMainWindow, show=True):
#     """ Opens the scanner. """
#     # If the scanner is not already running, create one
#     from .scanning.main import Scanner  # Inside to avoid circular import
#     if instances['scanner'] is None:
#         instances['scanner'] = Scanner(gui)
#         instances['scanner'].show()
#         instances['scanner'].activateWindow()
#         gui.activateWindow() # Put main window back to the front
#     # If the scanner is already running, just make as the front window
#     elif show:
#         instances['scanner'].setWindowState(
#             instances['scanner'].windowState()
#             & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
#         instances['scanner'].activateWindow()


# def clearScanner():
#     """ Clears the scanner instance reference when quitted """
#     instances['scanner'] = None


# def closeScanner():
#     if instances['scanner'] is not None:
#         instances['scanner'].close()
